// This file is part of the EonaCat project(s) which is released under the Apache License.
// See the LICENSE file or go to https://EonaCat.com/License for full license details.

using System;
using System.Collections.Generic;
using System.IO;
using System.Net;
using System.Net.Security;
using System.Net.Sockets;
using System.Text;
using System.Threading;

namespace EonaCat.Network
{
    /// <summary>
    /// Represents an HTTP connection.
    /// </summary>
    internal sealed class HttpConnection
    {
        private byte[] _buffer;
        private const int _bufferLength = 8192;
        private HttpListenerContext _context;
        private bool _contextRegistered;
        private StringBuilder _currentLine;
        private InputState _inputState;
        private RequestStream _inputStream;
        private HttpListener _lastListener;
        private LineState _lineState;
        private readonly EndPointListener _listener;
        private ResponseStream _outputStream;
        private int _position;
        private MemoryStream _requestBuffer;
        private Socket _socket;
        private readonly object _sync;
        private int _timeout;
        private readonly Dictionary<int, bool> _timeoutCanceled;
        private Timer _timer;

        /// <summary>
        /// Initializes a new instance of the <see cref="HttpConnection"/> class.
        /// </summary>
        /// <param name="socket">The socket associated with the connection.</param>
        /// <param name="listener">The endpoint listener.</param>
        internal HttpConnection(Socket socket, EndPointListener listener)
        {
            _socket = socket;
            _listener = listener;
            IsSecure = listener.IsSecure;

            var netStream = new NetworkStream(socket, false);
            if (IsSecure)
            {
                var conf = listener.SslConfiguration;
                var sslStream = new SslStream(netStream, false, conf.ClientCertificateValidationCallback);
                sslStream.AuthenticateAsServer(
                  conf.Certificate,
                  conf.IsClientCertificateRequired,
                  conf.SslProtocols,
                  conf.CheckForCertificateRevocation
                );

                Stream = sslStream;
            }
            else
            {
                Stream = netStream;
            }

            _sync = new object();
            _timeout = 90000; // 90k ms for first request, 15k ms from then on.
            _timeoutCanceled = new Dictionary<int, bool>();
            _timer = new Timer(onTimeout, this, Timeout.Infinite, Timeout.Infinite);

            init();
        }

        /// <summary>
        /// Gets a value indicating whether the connection is closed.
        /// </summary>
        public bool IsClosed => _socket == null;

        /// <summary>
        /// Gets a value indicating whether the connection is secure.
        /// </summary>
        public bool IsSecure { get; }

        /// <summary>
        /// Gets the local endpoint.
        /// </summary>
        public IPEndPoint LocalEndPoint => (IPEndPoint)_socket.LocalEndPoint;

        /// <summary>
        /// Gets the remote endpoint.
        /// </summary>
        public IPEndPoint RemoteEndPoint => (IPEndPoint)_socket.RemoteEndPoint;

        /// <summary>
        /// Gets or sets the number of reuses.
        /// </summary>
        public int Reuses { get; private set; }

        /// <summary>
        /// Gets the network stream associated with the connection.
        /// </summary>
        public Stream Stream { get; private set; }

        private void close()
        {
            lock (_sync)
            {
                if (_socket == null)
                {
                    return;
                }

                disposeTimer();
                disposeRequestBuffer();
                disposeStream();
                closeSocket();
            }

            unregisterContext();
            removeConnection();
        }

        private void closeSocket()
        {
            try
            {
                _socket.Shutdown(SocketShutdown.Both);
            }
            catch
            {
            }

            _socket.Close();
            _socket = null;
        }

        private void disposeRequestBuffer()
        {
            if (_requestBuffer == null)
            {
                return;
            }

            _requestBuffer.Dispose();
            _requestBuffer = null;
        }

        private void disposeStream()
        {
            if (Stream == null)
            {
                return;
            }

            _inputStream = null;
            _outputStream = null;

            Stream.Dispose();
            Stream = null;
        }

        private void disposeTimer()
        {
            if (_timer == null)
            {
                return;
            }

            try
            {
                _timer.Change(Timeout.Infinite, Timeout.Infinite);
            }
            catch
            {
            }

            _timer.Dispose();
            _timer = null;
        }

        private void init()
        {
            _context = new HttpListenerContext(this);
            _inputState = InputState.RequestLine;
            _inputStream = null;
            _lineState = LineState.None;
            _outputStream = null;
            _position = 0;
            _requestBuffer = new MemoryStream();
        }

        private static void onRead(IAsyncResult asyncResult)
        {
            var conn = (HttpConnection)asyncResult.AsyncState;
            if (conn._socket == null)
            {
                return;
            }

            lock (conn._sync)
            {
                if (conn._socket == null)
                {
                    return;
                }

                var nread = -1;
                var len = 0;
                try
                {
                    var current = conn.Reuses;
                    if (!conn._timeoutCanceled[current])
                    {
                        conn._timer.Change(Timeout.Infinite, Timeout.Infinite);
                        conn._timeoutCanceled[current] = true;
                    }

                    nread = conn.Stream.EndRead(asyncResult);
                    conn._requestBuffer.Write(conn._buffer, 0, nread);
                    len = (int)conn._requestBuffer.Length;
                }
                catch (Exception ex)
                {
                    if (conn._requestBuffer != null && conn._requestBuffer.Length > 0)
                    {
                        conn.SendError(ex.Message, 400);
                        return;
                    }

                    conn.close();
                    return;
                }

                if (nread <= 0)
                {
                    conn.close();
                    return;
                }

                if (conn.processInput(conn._requestBuffer.GetBuffer(), len))
                {
                    if (!conn._context.HasError)
                    {
                        conn._context.Request.FinishInitialization();
                    }

                    if (conn._context.HasError)
                    {
                        conn.SendError();
                        return;
                    }

                    HttpListener lsnr;
                    if (!conn._listener.TrySearchHttpListener(conn._context.Request.Url, out lsnr))
                    {
                        conn.SendError(null, 404);
                        return;
                    }

                    if (conn._lastListener != lsnr)
                    {
                        conn.removeConnection();
                        if (!lsnr.AddConnection(conn))
                        {
                            conn.close();
                            return;
                        }

                        conn._lastListener = lsnr;
                    }

                    conn._context.Listener = lsnr;
                    if (!conn._context.Authenticate())
                    {
                        return;
                    }

                    if (conn._context.Register())
                    {
                        conn._contextRegistered = true;
                    }

                    return;
                }

                conn.Stream.BeginRead(conn._buffer, 0, _bufferLength, onRead, conn);
            }
        }

        private static void onTimeout(object state)
        {
            var conn = (HttpConnection)state;
            var current = conn.Reuses;
            if (conn._socket == null)
            {
                return;
            }

            lock (conn._sync)
            {
                if (conn._socket == null)
                {
                    return;
                }

                if (conn._timeoutCanceled[current])
                {
                    return;
                }

                conn.SendError(null, 408);
            }
        }

        private bool processInput(byte[] data, int length)
        {
            _currentLine ??= new StringBuilder(64);

            var nread = 0;
            try
            {
                string line;
                while ((line = readLineFrom(data, _position, length, out nread)) != null)
                {
                    _position += nread;
                    if (line.Length == 0)
                    {
                        if (_inputState == InputState.RequestLine)
                        {
                            continue;
                        }

                        if (_position > 32768)
                        {
                            _context.ErrorMessage = "Headers too long";
                        }

                        _currentLine = null;
                        return true;
                    }

                    if (_inputState == InputState.RequestLine)
                    {
                        _context.Request.SetRequestLine(line);
                        _inputState = InputState.Headers;
                    }
                    else
                    {
                        _context.Request.AddHeader(line);
                    }

                    if (_context.HasError)
                    {
                        return true;
                    }
                }
            }
            catch (Exception ex)
            {
                _context.ErrorMessage = ex.Message;
                return true;
            }

            _position += nread;
            if (_position >= 32768)
            {
                _context.ErrorMessage = "Headers too long";
                return true;
            }

            return false;
        }

        private string readLineFrom(byte[] buffer, int offset, int length, out int read)
        {
            read = 0;

            for (var i = offset; i < length && _lineState != LineState.Lf; i++)
            {
                read++;

                var b = buffer[i];
                if (b == 13)
                {
                    _lineState = LineState.Cr;
                }
                else if (b == 10)
                {
                    _lineState = LineState.Lf;
                }
                else
                {
                    _currentLine.Append((char)b);
                }
            }

            if (_lineState != LineState.Lf)
            {
                return null;
            }

            var line = _currentLine.ToString();

            _currentLine.Length = 0;
            _lineState = LineState.None;

            return line;
        }

        private void removeConnection()
        {
            if (_lastListener != null)
            {
                _lastListener.RemoveConnection(this);
            }
            else
            {
                _listener.RemoveConnection(this);
            }
        }

        private void unregisterContext()
        {
            if (!_contextRegistered)
            {
                return;
            }

            _context.Unregister();
            _contextRegistered = false;
        }

        /// <summary>
        /// Closes the connection.
        /// </summary>
        /// <param name="force">True to force close, false otherwise.</param>
        internal void Close(bool force)
        {
            if (_socket == null)
            {
                return;
            }

            lock (_sync)
            {
                if (_socket == null)
                {
                    return;
                }

                if (!force)
                {
                    GetResponseStream().Close(false);
                    if (!_context.Response.CloseConnection && _context.Request.FlushInput())
                    {
                        // Don't close. Keep working.
                        Reuses++;
                        disposeRequestBuffer();
                        unregisterContext();
                        init();
                        BeginReadRequest();

                        return;
                    }
                }
                else if (_outputStream != null)
                {
                    _outputStream.Close(true);
                }

                close();
            }
        }

        /// <summary>
        /// Initiates reading the request.
        /// </summary>
        public void BeginReadRequest()
        {
            _buffer ??= new byte[_bufferLength];

            if (Reuses == 1)
            {
                _timeout = 15000;
            }

            try
            {
                _timeoutCanceled.Add(Reuses, false);
                _timer.Change(_timeout, Timeout.Infinite);
                Stream.BeginRead(_buffer, 0, _bufferLength, onRead, this);
            }
            catch
            {
                close();
            }
        }

        /// <summary>
        /// Closes the connection.
        /// </summary>
        public void Close()
        {
            Close(false);
        }

        /// <summary>
        /// Gets the request stream.
        /// </summary>
        /// <param name="contentLength">The length of the content.</param>
        /// <param name="chunked">True if chunked, false otherwise.</param>
        /// <returns>The request stream.</returns>
        public RequestStream GetRequestStream(long contentLength, bool chunked)
        {
            if (_inputStream != null || _socket == null)
            {
                return _inputStream;
            }

            lock (_sync)
            {
                if (_socket == null)
                {
                    return _inputStream;
                }

                var buff = _requestBuffer.GetBuffer();
                var len = (int)_requestBuffer.Length;
                disposeRequestBuffer();
                if (chunked)
                {
                    _context.Response.SendInChunks = true;
                    _inputStream =
                      new ChunkedRequestStream(Stream, buff, _position, len - _position, _context);
                }
                else
                {
                    _inputStream =
                      new RequestStream(Stream, buff, _position, len - _position, contentLength);
                }

                return _inputStream;
            }
        }

        /// <summary>
        /// Gets the response stream.
        /// </summary>
        /// <returns>The response stream.</returns>
        public ResponseStream GetResponseStream()
        {
            if (_outputStream != null || _socket == null)
            {
                return _outputStream;
            }

            lock (_sync)
            {
                if (_socket == null)
                {
                    return _outputStream;
                }

                var lsnr = _context.Listener;
                var ignore = lsnr == null || lsnr.IgnoreWriteExceptions;
                _outputStream = new ResponseStream(Stream, _context.Response, ignore);

                return _outputStream;
            }
        }

        /// <summary>
        /// Sends an error response.
        /// </summary>
        public void SendError()
        {
            SendError(_context.ErrorMessage, _context.ErrorStatus);
        }

        /// <summary>
        /// Sends an error response with the specified message and status code.
        /// </summary>
        /// <param name="message">The error message.</param>
        /// <param name="status">The HTTP status code.</param>
        public void SendError(string message, int status)
        {
            if (_socket == null)
            {
                return;
            }

            lock (_sync)
            {
                if (_socket == null)
                {
                    return;
                }

                try
                {
                    var res = _context.Response;
                    res.StatusCode = status;
                    res.ContentType = "text/html";

                    var content = new StringBuilder(64);
                    content.AppendFormat("<html><head><title>EonaCat.Network Error</title></head><body><h1>{0} {1}", status, res.StatusDescription);
                    if (message != null && message.Length > 0)
                    {
                        content.AppendFormat(" ({0})</h1></body></html>", message);
                    }
                    else
                    {
                        content.Append("</h1></body></html>");
                    }

                    var enc = Encoding.UTF8;
                    var entity = enc.GetBytes(content.ToString());
                    res.ContentEncoding = enc;
                    res.ContentLength64 = entity.LongLength;

                    res.Close(entity, true);
                }
                catch
                {
                    Close(true);
                }
            }
        }
    }
}