// This file is part of the EonaCat project(s) which is released under the Apache License.
// See the LICENSE file or go to https://EonaCat.com/License for full license details.

using System;
using System.Collections;
using System.Collections.Generic;
using System.Security.Principal;

namespace EonaCat.Network
{
    /// <summary>
    /// Represents an HTTP listener that listens for incoming requests.
    /// </summary>
    public sealed class HttpListener : IDisposable
    {
        private AuthenticationSchemes _authSchemes;
        private Func<HttpListenerRequest, AuthenticationSchemes> _authSchemeSelector;
        private string _certFolderPath;
        private readonly Dictionary<HttpConnection, HttpConnection> _connections;
        private readonly object _connectionsSync;
        private readonly List<HttpListenerContext> _ctxQueue;
        private readonly object _ctxQueueSync;
        private readonly Dictionary<HttpListenerContext, HttpListenerContext> _ctxRegistry;
        private readonly object _ctxRegistrySync;
        private static readonly string _defaultRealm;
        private bool _ignoreWriteExceptions;
        private volatile bool _listening;
        private readonly HttpListenerPrefixCollection _prefixes;
        private string _realm;
        private SSLConfigurationServer _sslConfig;
        private Func<IIdentity, NetworkCredential> _userCredFinder;
        private readonly List<HttpListenerAsyncResult> _waitQueue;
        private readonly object _waitQueueSync;

        static HttpListener()
        {
            _defaultRealm = "SECRET AREA";
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="HttpListener"/> class.
        /// </summary>
        public HttpListener()
        {
            _authSchemes = AuthenticationSchemes.Anonymous;

            _connections = new Dictionary<HttpConnection, HttpConnection>();
            _connectionsSync = ((ICollection)_connections).SyncRoot;

            _ctxQueue = new List<HttpListenerContext>();
            _ctxQueueSync = ((ICollection)_ctxQueue).SyncRoot;

            _ctxRegistry = new Dictionary<HttpListenerContext, HttpListenerContext>();
            _ctxRegistrySync = ((ICollection)_ctxRegistry).SyncRoot;

            _prefixes = new HttpListenerPrefixCollection(this);

            _waitQueue = new List<HttpListenerAsyncResult>();
            _waitQueueSync = ((ICollection)_waitQueue).SyncRoot;
        }

        internal bool IsDisposed { get; private set; }

        internal bool ReuseAddress { get; set; }

        /// <summary>
        /// Gets or sets the authentication schemes used by this listener.
        /// </summary>
        public AuthenticationSchemes AuthenticationSchemes
        {
            get
            {
                CheckDisposed();
                return _authSchemes;
            }

            set
            {
                CheckDisposed();
                _authSchemes = value;
            }
        }

        public Func<HttpListenerRequest, AuthenticationSchemes> AuthenticationSchemeSelector
        {
            get
            {
                CheckDisposed();
                return _authSchemeSelector;
            }

            set
            {
                CheckDisposed();
                _authSchemeSelector = value;
            }
        }

        public string CertificateFolderPath
        {
            get
            {
                CheckDisposed();
                return _certFolderPath;
            }

            set
            {
                CheckDisposed();
                _certFolderPath = value;
            }
        }

        public bool IgnoreWriteExceptions
        {
            get
            {
                CheckDisposed();
                return _ignoreWriteExceptions;
            }

            set
            {
                CheckDisposed();
                _ignoreWriteExceptions = value;
            }
        }

        public bool IsListening => _listening;

        public static bool IsSupported => true;

        public HttpListenerPrefixCollection Prefixes
        {
            get
            {
                CheckDisposed();
                return _prefixes;
            }
        }

        public string Realm
        {
            get
            {
                CheckDisposed();
                return _realm;
            }

            set
            {
                CheckDisposed();
                _realm = value;
            }
        }

        public SSLConfigurationServer SslConfiguration
        {
            get
            {
                CheckDisposed();
                return _sslConfig ??= new SSLConfigurationServer();
            }

            set
            {
                CheckDisposed();
                _sslConfig = value;
            }
        }

        public bool UnsafeConnectionNtlmAuthentication
        {
            get
            {
                throw new NotSupportedException();
            }

            set
            {
                throw new NotSupportedException();
            }
        }

        public Func<IIdentity, NetworkCredential> UserCredentialsFinder
        {
            get
            {
                CheckDisposed();
                return _userCredFinder;
            }

            set
            {
                CheckDisposed();
                _userCredFinder = value;
            }
        }

        private void cleanupConnections()
        {
            HttpConnection[] conns = null;
            lock (_connectionsSync)
            {
                if (_connections.Count == 0)
                {
                    return;
                }

                // Need to copy this since closing will call the RemoveConnection method.
                var keys = _connections.Keys;
                conns = new HttpConnection[keys.Count];
                keys.CopyTo(conns, 0);
                _connections.Clear();
            }

            for (var i = conns.Length - 1; i >= 0; i--)
            {
                conns[i].Close(true);
            }
        }

        private void cleanupContextQueue(bool sendServiceUnavailable)
        {
            HttpListenerContext[] ctxs = null;
            lock (_ctxQueueSync)
            {
                if (_ctxQueue.Count == 0)
                {
                    return;
                }

                ctxs = _ctxQueue.ToArray();
                _ctxQueue.Clear();
            }

            if (!sendServiceUnavailable)
            {
                return;
            }

            foreach (var ctx in ctxs)
            {
                var res = ctx.Response;
                res.StatusCode = (int)HttpStatusCode.ServiceUnavailable;
                res.Close();
            }
        }

        private void cleanupContextRegistry()
        {
            HttpListenerContext[] ctxs = null;
            lock (_ctxRegistrySync)
            {
                if (_ctxRegistry.Count == 0)
                {
                    return;
                }

                // Need to copy this since closing will call the UnregisterContext method.
                var keys = _ctxRegistry.Keys;
                ctxs = new HttpListenerContext[keys.Count];
                keys.CopyTo(ctxs, 0);
                _ctxRegistry.Clear();
            }

            for (var i = ctxs.Length - 1; i >= 0; i--)
            {
                ctxs[i].Connection.Close(true);
            }
        }

        private void cleanupWaitQueue(Exception exception)
        {
            HttpListenerAsyncResult[] aress = null;
            lock (_waitQueueSync)
            {
                if (_waitQueue.Count == 0)
                {
                    return;
                }

                aress = _waitQueue.ToArray();
                _waitQueue.Clear();
            }

            foreach (var ares in aress)
            {
                ares.Complete(exception);
            }
        }

        private void close(bool force)
        {
            if (_listening)
            {
                _listening = false;
                EndPointManager.RemoveListener(this);
            }

            lock (_ctxRegistrySync)
            {
                cleanupContextQueue(!force);
            }

            cleanupContextRegistry();
            cleanupConnections();
            cleanupWaitQueue(new ObjectDisposedException(GetType().ToString()));

            IsDisposed = true;
        }

        private HttpListenerAsyncResult getAsyncResultFromQueue()
        {
            if (_waitQueue.Count == 0)
            {
                return null;
            }

            var ares = _waitQueue[0];
            _waitQueue.RemoveAt(0);

            return ares;
        }

        private HttpListenerContext getContextFromQueue()
        {
            if (_ctxQueue.Count == 0)
            {
                return null;
            }

            var ctx = _ctxQueue[0];
            _ctxQueue.RemoveAt(0);

            return ctx;
        }

        internal bool AddConnection(HttpConnection connection)
        {
            if (!_listening)
            {
                return false;
            }

            lock (_connectionsSync)
            {
                if (!_listening)
                {
                    return false;
                }

                _connections[connection] = connection;
                return true;
            }
        }

        internal HttpListenerAsyncResult BeginGetContext(HttpListenerAsyncResult asyncResult)
        {
            lock (_ctxRegistrySync)
            {
                if (!_listening)
                {
                    throw new HttpListenerException(995);
                }

                var ctx = getContextFromQueue();
                if (ctx == null)
                {
                    _waitQueue.Add(asyncResult);
                }
                else
                {
                    asyncResult.Complete(ctx, true);
                }

                return asyncResult;
            }
        }

        internal void CheckDisposed()
        {
            if (IsDisposed)
            {
                throw new ObjectDisposedException(GetType().ToString());
            }
        }

        internal string GetRealm()
        {
            var realm = _realm;
            return realm != null && realm.Length > 0 ? realm : _defaultRealm;
        }

        internal Func<IIdentity, NetworkCredential> GetUserCredentialsFinder()
        {
            return _userCredFinder;
        }

        internal bool RegisterContext(HttpListenerContext context)
        {
            if (!_listening)
            {
                return false;
            }

            lock (_ctxRegistrySync)
            {
                if (!_listening)
                {
                    return false;
                }

                _ctxRegistry[context] = context;

                var ares = getAsyncResultFromQueue();
                if (ares == null)
                {
                    _ctxQueue.Add(context);
                }
                else
                {
                    ares.Complete(context);
                }

                return true;
            }
        }

        internal void RemoveConnection(HttpConnection connection)
        {
            lock (_connectionsSync)
            {
                _connections.Remove(connection);
            }
        }

        internal AuthenticationSchemes SelectAuthenticationScheme(HttpListenerRequest request)
        {
            var selector = _authSchemeSelector;
            if (selector == null)
            {
                return _authSchemes;
            }

            try
            {
                return selector(request);
            }
            catch
            {
                return AuthenticationSchemes.None;
            }
        }

        internal void UnregisterContext(HttpListenerContext context)
        {
            lock (_ctxRegistrySync)
            {
                _ctxRegistry.Remove(context);
            }
        }

        /// <summary>
        /// Aborts the listener and releases all resources associated with it.
        /// </summary>
        public void Abort()
        {
            if (IsDisposed)
            {
                return;
            }

            close(true);
        }

        /// <summary>
        /// Begins asynchronously getting an HTTP context from the listener.
        /// </summary>
        /// <param name="callback">The method to call when the operation completes.</param>
        /// <param name="state">A user-defined object that contains information about the asynchronous operation.</param>
        /// <returns>An <see cref="IAsyncResult"/> that represents the asynchronous operation.</returns>
        public IAsyncResult BeginGetContext(AsyncCallback callback, object state)
        {
            CheckDisposed();
            if (_prefixes.Count == 0)
            {
                throw new InvalidOperationException("The listener has no URI prefix on which listens.");
            }

            if (!_listening)
            {
                throw new InvalidOperationException("The listener hasn't been started.");
            }

            return BeginGetContext(new HttpListenerAsyncResult(callback, state));
        }

        /// <summary>
        /// Closes the listener.
        /// </summary>
        public void Close()
        {
            if (IsDisposed)
            {
                return;
            }

            close(false);
        }

        /// <summary>
        /// Ends an asynchronous operation to get an HTTP context from the listener.
        /// </summary>
        /// <param name="asyncResult">The reference to the pending asynchronous request to finish.</param>
        /// <returns>An <see cref="HttpListenerContext"/> that represents the context of the asynchronous operation.</returns>
        public HttpListenerContext EndGetContext(IAsyncResult asyncResult)
        {
            CheckDisposed();
            if (asyncResult == null)
            {
                throw new ArgumentNullException(nameof(asyncResult));
            }

            if (asyncResult is not HttpListenerAsyncResult ares)
            {
                throw new ArgumentException("A wrong IAsyncResult.", nameof(asyncResult));
            }

            if (ares.EndCalled)
            {
                throw new InvalidOperationException("This IAsyncResult cannot be reused.");
            }

            ares.EndCalled = true;
            if (!ares.IsCompleted)
            {
                ares.AsyncWaitHandle.WaitOne();
            }

            return ares.GetContext(); // This may throw an exception.
        }

        /// <summary>
        /// Gets the next available HTTP context from the listener.
        /// </summary>
        /// <returns>An <see cref="HttpListenerContext"/> that represents the context of the HTTP request.</returns>
        public HttpListenerContext GetContext()
        {
            CheckDisposed();
            if (_prefixes.Count == 0)
            {
                throw new InvalidOperationException("The listener has no URI prefix on which listens.");
            }

            if (!_listening)
            {
                throw new InvalidOperationException("The listener hasn't been started.");
            }

            var ares = BeginGetContext(new HttpListenerAsyncResult(null, null));
            ares.InGet = true;

            return EndGetContext(ares);
        }

        /// <summary>
        /// Starts listening for incoming requests.
        /// </summary>
        public void Start()
        {
            CheckDisposed();
            if (_listening)
            {
                return;
            }

            EndPointManager.AddListener(this);
            _listening = true;
        }

        /// <summary>
        /// Stops listening for incoming requests.
        /// </summary>
        public void Stop()
        {
            CheckDisposed();
            if (!_listening)
            {
                return;
            }

            _listening = false;
            EndPointManager.RemoveListener(this);

            lock (_ctxRegistrySync)
            {
                cleanupContextQueue(true);
            }

            cleanupContextRegistry();
            cleanupConnections();
            cleanupWaitQueue(new HttpListenerException(995, "The listener is closed."));
        }

        /// <summary>
        /// Disposes of the resources used by the <see cref="HttpListener"/>.
        /// </summary>
        void IDisposable.Dispose()
        {
            if (IsDisposed)
            {
                return;
            }

            close(true);
        }
    }
}