// This file is part of the EonaCat project(s) which is released under the Apache License.
// See the LICENSE file or go to https://EonaCat.com/License for full license details.

using System;
using System.Net.Sockets;
using System.Security.Principal;
using System.Threading;

namespace EonaCat.Network
{
    public class WSServer
    {
        public bool IsConsoleLoggingEnabled { get; set; }
        public bool IsLoggingEnabled { get; set; }
        private bool _allowForwardedRequest;
        private AuthenticationSchemes _authSchemes;
        private static readonly string _defaultRealm;
        private bool _dnsStyle;
        private string _hostname;
        private TcpListener _listener;
        private string _realm;
        private string _realmInUse;
        private Thread _receiveThread;
        private bool _reuseAddress;
        private SSLConfigurationServer _sslConfig;
        private SSLConfigurationServer _sslConfigInUse;
        private volatile ServerState _state;
        private object _sync;
        private Func<IIdentity, NetworkCredential> _userCredentialsFinder;

        static WSServer()
        {
            _defaultRealm = "SECRET AREA";
        }

        public WSServer()
        {
            var addr = System.Net.IPAddress.Any;
            init(addr.ToString(), addr, 80, false);
        }

        public WSServer(int port)
          : this(port, port == 443)
        {
        }

        public WSServer(string url)
        {
            if (url == null)
            {
                throw new ArgumentNullException(nameof(url));
            }

            if (url.Length == 0)
            {
                throw new ArgumentException("An empty string.", nameof(url));
            }

            Uri uri;
            string msg;
            if (!tryCreateUri(url, out uri, out msg))
            {
                throw new ArgumentException(msg, nameof(url));
            }

            var host = uri.DnsSafeHost;

            var addr = host.ToIPAddress();
            if (addr == null)
            {
                msg = "The host part could not be converted to an IP address.";
                throw new ArgumentException(msg, nameof(url));
            }

            if (!addr.IsLocal())
            {
                msg = "The IP address of the host is not a local IP address.";
                throw new ArgumentException(msg, nameof(url));
            }

            init(host, addr, uri.Port, uri.Scheme == "wss");
        }

        public WSServer(int port, bool secure)
        {
            if (!port.IsPortNumber())
            {
                var msg = "Less than 1 or greater than 65535.";
                throw new ArgumentOutOfRangeException(nameof(port), msg);
            }

            var addr = System.Net.IPAddress.Any;
            init(addr.ToString(), addr, port, secure);
        }

        public WSServer(System.Net.IPAddress address, int port)
          : this(address, port, port == 443)
        {
        }

        public WSServer(System.Net.IPAddress address, int port, bool secure)
        {
            if (address == null)
            {
                throw new ArgumentNullException(nameof(address));
            }

            if (!address.IsLocal())
            {
                throw new ArgumentException("Not a local IP address.", nameof(address));
            }

            if (!port.IsPortNumber())
            {
                var msg = "Less than 1 or greater than 65535.";
                throw new ArgumentOutOfRangeException(nameof(port), msg);
            }

            init(address.ToString(), address, port, secure);
        }

        public System.Net.IPAddress Address { get; private set; }

        public bool AllowForwardedRequest
        {
            get
            {
                return _allowForwardedRequest;
            }

            set
            {
                string msg;
                if (!CanSet(out msg))
                {
                    Logger.Warning(msg);
                    return;
                }

                lock (_sync)
                {
                    if (!CanSet(out msg))
                    {
                        Logger.Warning(msg);
                        return;
                    }

                    _allowForwardedRequest = value;
                }
            }
        }

        public AuthenticationSchemes AuthenticationSchemes
        {
            get
            {
                return _authSchemes;
            }

            set
            {
                string msg;
                if (!CanSet(out msg))
                {
                    Logger.Warning(msg);
                    return;
                }

                lock (_sync)
                {
                    if (!CanSet(out msg))
                    {
                        Logger.Warning(msg);
                        return;
                    }

                    _authSchemes = value;
                }
            }
        }

        public bool IsListening => _state == ServerState.Start;

        public bool IsSecure { get; private set; }

        /// <summary>
        /// Determines if sessions need to be removed automatically
        /// </summary>
        public bool AutoCleanSessions
        {
            get
            {
                return Endpoints.AutoCleanSessions;
            }

            set
            {
                Endpoints.AutoCleanSessions = value;
            }
        }

        public int Port { get; private set; }

        public string Realm
        {
            get
            {
                return _realm;
            }

            set
            {
                string msg;
                if (!CanSet(out msg))
                {
                    Logger.Warning(msg);
                    return;
                }

                lock (_sync)
                {
                    if (!CanSet(out msg))
                    {
                        Logger.Warning(msg);
                        return;
                    }

                    _realm = value;
                }
            }
        }

        public bool ReuseAddress
        {
            get
            {
                return _reuseAddress;
            }

            set
            {
                string msg;
                if (!CanSet(out msg))
                {
                    Logger.Warning(msg);
                    return;
                }

                lock (_sync)
                {
                    if (!CanSet(out msg))
                    {
                        Logger.Warning(msg);
                        return;
                    }

                    _reuseAddress = value;
                }
            }
        }

        public SSLConfigurationServer SslConfiguration
        {
            get
            {
                if (!IsSecure)
                {
                    var msg = "This instance does not provide secure connections.";
                    throw new InvalidOperationException(msg);
                }

                return GetSslConfiguration();
            }
        }

        public Func<IIdentity, NetworkCredential> FindCredentials
        {
            get
            {
                return _userCredentialsFinder;
            }

            set
            {
                string message;
                if (!CanSet(out message))
                {
                    Logger.Warning(message);
                    return;
                }

                lock (_sync)
                {
                    if (!CanSet(out message))
                    {
                        Logger.Warning(message);
                        return;
                    }

                    _userCredentialsFinder = value;
                }
            }
        }

        public TimeSpan WaitTime
        {
            get
            {
                return Endpoints.WaitTime;
            }

            set
            {
                Endpoints.WaitTime = value;
            }
        }

        public WSEndpointManager Endpoints { get; private set; }

        private void abort()
        {
            lock (_sync)
            {
                if (_state != ServerState.Start)
                {
                    return;
                }

                _state = ServerState.ShuttingDown;
            }

            try
            {
                try
                {
                    _listener.Stop();
                }
                finally
                {
                    Endpoints.Stop(1006, string.Empty);
                }
            }
            catch
            {
            }

            _state = ServerState.Stop;
        }

        private bool CanSet(out string message)
        {
            message = null;

            if (_state == ServerState.Start)
            {
                message = "The server has already started.";
                return false;
            }

            if (_state == ServerState.ShuttingDown)
            {
                message = "The server is shutting down.";
                return false;
            }

            return true;
        }

        private bool CheckHostNameForRequest(string name)
        {
            return !_dnsStyle
                   || Uri.CheckHostName(name) != UriHostNameType.Dns
                   || name == _hostname;
        }

        private static bool CheckSslConfiguration(
          SSLConfigurationServer configuration, out string message
        )
        {
            message = null;

            if (configuration.Certificate == null)
            {
                message = "There is no server certificate for secure connections.";
                return false;
            }

            return true;
        }

        private string GetRealm()
        {
            var realm = _realm;
            return realm != null && realm.Length > 0 ? realm : _defaultRealm;
        }

        private SSLConfigurationServer GetSslConfiguration()
        {
            _sslConfig ??= new SSLConfigurationServer();

            return _sslConfig;
        }

        private void init(
          string hostname, System.Net.IPAddress address, int port, bool secure
        )
        {
            _hostname = hostname;
            Address = address;
            Port = port;
            IsSecure = secure;

            _authSchemes = AuthenticationSchemes.Anonymous;
            _dnsStyle = Uri.CheckHostName(hostname) == UriHostNameType.Dns;
            _listener = new TcpListener(address, port);
            Endpoints = new WSEndpointManager();
            _sync = new object();
        }

        private void processRequest(TcpListenerWebSocketContext context)
        {
            var uri = context.RequestUri;
            if (uri == null)
            {
                context.Close(HttpStatusCode.BadRequest);
                return;
            }

            if (!_allowForwardedRequest)
            {
                if (uri.Port != Port)
                {
                    context.Close(HttpStatusCode.BadRequest);
                    return;
                }

                if (!CheckHostNameForRequest(uri.DnsSafeHost))
                {
                    context.Close(HttpStatusCode.NotFound);
                    return;
                }
            }

            WSEndpointHost host;
            if (!Endpoints.InternalTryGetServiceHost(uri.AbsolutePath, out host))
            {
                context.Close(HttpStatusCode.NotImplemented);
                return;
            }

            host.StartSession(context);
        }

        private void receiveRequest()
        {
            while (true)
            {
                TcpClient cl = null;
                try
                {
                    cl = _listener.AcceptTcpClient();
                    ThreadPool.QueueUserWorkItem(
                      state =>
                      {
                          try
                          {
                              var ctx = new TcpListenerWebSocketContext(
                                cl, null, IsSecure, _sslConfigInUse);

                              if (!ctx.Authenticate(_authSchemes, _realmInUse, _userCredentialsFinder))
                              {
                                  return;
                              }

                              processRequest(ctx);
                          }
                          catch (Exception ex)
                          {
                              Logger.Error(ex.Message);
                              Logger.Debug(ex.ToString());

                              cl.Close();
                          }
                      }
                    );
                }
                catch (SocketException ex)
                {
                    if (_state == ServerState.ShuttingDown)
                    {
                        Logger.Info("The underlying listener is stopped.");
                        break;
                    }

                    Logger.Error(ex.Message);
                    Logger.Debug(ex.ToString());

                    break;
                }
                catch (Exception ex)
                {
                    Logger.Error(ex.Message);
                    Logger.Debug(ex.ToString());

                    if (cl != null)
                    {
                        cl.Close();
                    }

                    break;
                }
            }

            if (_state != ServerState.ShuttingDown)
            {
                abort();
            }
        }

        private void start(SSLConfigurationServer sslConfig)
        {
            Logger.IsLoggingEnabled = IsLoggingEnabled;
            Logger.DisableConsole = !IsConsoleLoggingEnabled;

            if (_state == ServerState.Start)
            {
                Logger.Info("The server has already started.");
                return;
            }

            if (_state == ServerState.ShuttingDown)
            {
                Logger.Warning("The server is shutting down.");
                return;
            }

            lock (_sync)
            {
                if (_state == ServerState.Start)
                {
                    Logger.Info("The server has already started.");
                    return;
                }

                if (_state == ServerState.ShuttingDown)
                {
                    Logger.Warning("The server is shutting down.");
                    return;
                }

                _sslConfigInUse = sslConfig;
                _realmInUse = GetRealm();

                Endpoints.Start();
                try
                {
                    startReceiving();
                }
                catch
                {
                    Endpoints.Stop(1011, string.Empty);
                    throw;
                }

                _state = ServerState.Start;
            }
        }

        private void startReceiving()
        {
            if (_reuseAddress)
            {
                _listener.Server.SetSocketOption(
                  SocketOptionLevel.Socket, SocketOptionName.ReuseAddress, true
                );
            }

            try
            {
                _listener.Start();
            }
            catch (Exception ex)
            {
                var msg = "The underlying listener has failed to start.";
                throw new InvalidOperationException(msg, ex);
            }

            _receiveThread = new Thread(new ThreadStart(receiveRequest));
            _receiveThread.IsBackground = true;
            _receiveThread.Start();
        }

        private void stop(ushort code, string reason)
        {
            if (_state == ServerState.Ready)
            {
                Logger.Info("The server is not started.");
                return;
            }

            if (_state == ServerState.ShuttingDown)
            {
                Logger.Info("The server is shutting down.");
                return;
            }

            if (_state == ServerState.Stop)
            {
                Logger.Info("The server has already stopped.");
                return;
            }

            lock (_sync)
            {
                if (_state == ServerState.ShuttingDown)
                {
                    Logger.Info("The server is shutting down.");
                    return;
                }

                if (_state == ServerState.Stop)
                {
                    Logger.Info("The server has already stopped.");
                    return;
                }

                _state = ServerState.ShuttingDown;
            }

            try
            {
                var threw = false;
                try
                {
                    stopReceiving(5000);
                }
                catch
                {
                    threw = true;
                    throw;
                }
                finally
                {
                    try
                    {
                        Endpoints.Stop(code, reason);
                    }
                    catch
                    {
                        if (!threw)
                        {
                            throw;
                        }
                    }
                }
            }
            finally
            {
                _state = ServerState.Stop;
            }
        }

        private void stopReceiving(int millisecondsTimeout)
        {
            try
            {
                _listener.Stop();
            }
            catch (Exception ex)
            {
                var msg = "The underlying listener has failed to stop.";
                throw new InvalidOperationException(msg, ex);
            }

            _receiveThread.Join(millisecondsTimeout);
        }

        private static bool tryCreateUri(
          string uriString, out Uri result, out string message
        )
        {
            if (!uriString.TryCreateWebSocketUri(out result, out message))
            {
                return false;
            }

            if (result.PathAndQuery != "/")
            {
                result = null;
                message = "It includes either or both path and query components.";

                return false;
            }

            return true;
        }

        public void AddEndpoint<TEndpoint>(string path) where TEndpoint : WSEndpoint, new()
        {
            Endpoints.AddService<TEndpoint>(path, null);
        }

        public void AddEndpoint<TEndpoint>(string path, Action<TEndpoint> initializer) where TEndpoint : WSEndpoint, new()
        {
            Endpoints.AddService(path, initializer);
        }

        public bool RemoveEndpoint(string path)
        {
            return Endpoints.RemoveService(path);
        }

        public void Start()
        {
            SSLConfigurationServer sslConfig = null;

            if (IsSecure)
            {
                sslConfig = new SSLConfigurationServer(GetSslConfiguration());

                string message;
                if (!CheckSslConfiguration(sslConfig, out message))
                {
                    throw new InvalidOperationException(message);
                }
            }

            start(sslConfig);
        }

        public void Stop()
        {
            stop(1005, string.Empty);
        }

        public void Stop(ushort code, string reason)
        {
            if (!code.IsCloseStatusCode())
            {
                var msg = "Less than 1000 or greater than 4999.";
                throw new ArgumentOutOfRangeException(nameof(code), msg);
            }

            if (code == 1010)
            {
                var msg = "1010 cannot be used.";
                throw new ArgumentException(msg, nameof(code));
            }

            if (!reason.IsNullOrEmpty())
            {
                if (code == 1005)
                {
                    var msg = "1005 cannot be used.";
                    throw new ArgumentException(msg, nameof(code));
                }

                byte[] bytes;
                if (!reason.TryGetUTF8EncodedBytes(out bytes))
                {
                    var msg = "It could not be UTF-8-encoded.";
                    throw new ArgumentException(msg, nameof(reason));
                }

                if (bytes.Length > 123)
                {
                    var msg = "Its size is greater than 123 bytes.";
                    throw new ArgumentOutOfRangeException(nameof(reason), msg);
                }
            }

            stop(code, reason);
        }

        public void Stop(CloseStatusCode code, string reason)
        {
            if (code == CloseStatusCode.MandatoryExtension)
            {
                var msg = "MandatoryExtension cannot be used.";
                throw new ArgumentException(msg, nameof(code));
            }

            if (!reason.IsNullOrEmpty())
            {
                if (code == CloseStatusCode.NoStatus)
                {
                    var msg = "NoStatus cannot be used.";
                    throw new ArgumentException(msg, nameof(code));
                }

                byte[] bytes;
                if (!reason.TryGetUTF8EncodedBytes(out bytes))
                {
                    var msg = "It could not be UTF-8-encoded.";
                    throw new ArgumentException(msg, nameof(reason));
                }

                if (bytes.Length > 123)
                {
                    var msg = "Its size is greater than 123 bytes.";
                    throw new ArgumentOutOfRangeException(nameof(reason), msg);
                }
            }

            stop((ushort)code, reason);
        }
    }
}