// This file is part of the EonaCat project(s) which is released under the Apache License.
// See the LICENSE file or go to https://EonaCat.com/License for full license details.

using System;
using System.Collections;
using System.Collections.Generic;
using System.IO;
using System.Net.Security;
using System.Net.Sockets;
using System.Security.Cryptography;
using System.Text;
using System.Threading;
using System.Threading.Tasks;

namespace EonaCat.Network
{
    public class WebSocket : IDisposable
    {
        private AuthenticationChallenge _authChallenge;
        private string _base64Key;
        private readonly bool _client;
        private Action _closeContext;
        private CompressionMethod _compression;
        private WebSocketContext _context;
        private bool _isRedirectionEnabled;
        private string _extensions;
        private bool _extensionsRequested;
        private object _forMessageEventQueue;
        private object _forPing;
        private object _forSend;
        private object _forState;
        private MemoryStream _fragmentsBuffer;
        private bool _fragmentsCompressed;
        private Opcode _fragmentsOpcode;
        private const string _guid = "258EAFA5-E914-47DA-95CA-C5AB0DC85B11";
        private bool _inContinuation;
        private volatile bool _inMessage;
        private static readonly int _maxRetryCountForConnect;
        private readonly Action<MessageEventArgs> _message;
        private Queue<MessageEventArgs> _messageEventQueue;
        private uint _nonceCount;
        private string _origin;
        private ManualResetEvent _pongReceived;
        private bool _preAuth;
        private string _protocol;
        private readonly string[] _protocols;
        private bool _protocolsRequested;
        private NetworkCredential _proxyCredentials;
        private Uri _proxyUri;
        private volatile WebSocketState _readyState;
        private ManualResetEvent _receivingExited;
        private int _retryCountForConnect;
        private SSLConfigurationClient _sslConfig;
        private Stream _stream;
        private TcpClient _tcpClient;
        private Uri _uri;
        private const string _version = "13";
        private TimeSpan _waitTime;

        internal static readonly byte[] EmptyBytes;

        public static int FragmentLength;

        internal static readonly RandomNumberGenerator RandomNumber;

        static WebSocket()
        {
            _maxRetryCountForConnect = 10;
            EmptyBytes = new byte[0];
            FragmentLength = 1016;
            RandomNumber = new RNGCryptoServiceProvider();
        }

        // As server
        internal WebSocket(HttpListenerWebSocketContext context, string protocol)
        {
            _context = context;
            _protocol = protocol;

            _closeContext = context.Close;
            _message = messages;
            IsSecure = context.IsSecureConnection;
            _stream = context.Stream;
            _waitTime = TimeSpan.FromSeconds(1);

            init();
        }

        // As server
        internal WebSocket(TcpListenerWebSocketContext context, string protocol)
        {
            _context = context;
            _protocol = protocol;

            _closeContext = context.Close;
            _message = messages;
            IsSecure = context.IsSecureConnection;
            _stream = context.Stream;
            _waitTime = TimeSpan.FromSeconds(1);

            init();
        }

        public WebSocket(string url, params string[] protocols)
        {
            if (url == null)
            {
                throw new ArgumentNullException(nameof(url));
            }

            if (url.Length == 0)
            {
                throw new ArgumentException("An empty string.", nameof(url));
            }

            string msg;
            if (!url.TryCreateWebSocketUri(out _uri, out msg))
            {
                throw new ArgumentException(msg, nameof(url));
            }

            if (protocols != null && protocols.Length > 0)
            {
                if (!checkProtocols(protocols, out msg))
                {
                    throw new ArgumentException(msg, nameof(protocols));
                }

                _protocols = protocols;
            }

            _base64Key = CreateBase64Key();
            _client = true;
            _message = messagec;
            IsSecure = _uri.Scheme == "wss";
            _waitTime = TimeSpan.FromSeconds(5);

            init();
        }

        internal CookieCollection CookieCollection { get; private set; }

        // As server
        internal Func<WebSocketContext, string> CustomHandshakeRequestChecker { get; set; }

        internal bool HasMessage
        {
            get
            {
                lock (_forMessageEventQueue)
                {
                    return _messageEventQueue.Count > 0;
                }
            }
        }

        // As server
        internal bool IgnoreExtensions { get; set; }

        internal bool IsConnected => _readyState == WebSocketState.Open || _readyState == WebSocketState.Closing;

        public CompressionMethod Compression
        {
            get
            {
                return _compression;
            }

            set
            {
                string msg = null;

                if (!_client)
                {
                    msg = "The set operation cannot be used by servers.";
                    throw new InvalidOperationException(msg);
                }

                if (!canSet(out msg))
                {
                    Logger.Warning(msg);
                    return;
                }

                lock (_forState)
                {
                    if (!canSet(out msg))
                    {
                        Logger.Warning(msg);
                        return;
                    }

                    _compression = value;
                }
            }
        }

        public IEnumerable<Cookie> Cookies
        {
            get
            {
                lock (CookieCollection.SyncRoot)
                {
                    foreach (Cookie cookie in CookieCollection)
                    {
                        yield return cookie;
                    }
                }
            }
        }

        public NetworkCredential Credentials { get; private set; }

        public bool CallMessageOnPing { get; set; }

        public bool IsRedirectionEnabled
        {
            get
            {
                return _isRedirectionEnabled;
            }

            set
            {
                string msg = null;

                if (!_client)
                {
                    msg = "The set operation cannot be used by servers.";
                    throw new InvalidOperationException(msg);
                }

                if (!canSet(out msg))
                {
                    Logger.Warning(msg);
                    return;
                }

                lock (_forState)
                {
                    if (!canSet(out msg))
                    {
                        Logger.Warning(msg);
                        return;
                    }

                    _isRedirectionEnabled = value;
                }
            }
        }

        public string Extensions => _extensions ?? string.Empty;

        public bool IsAlive => ping(EmptyBytes);

        public bool IsSecure { get; private set; }

        public string Origin
        {
            get
            {
                return _origin;
            }

            set
            {
                string msg = null;

                if (!_client)
                {
                    msg = "This instance is not a client.";
                    throw new InvalidOperationException(msg);
                }

                if (!value.IsNullOrEmpty())
                {
                    Uri uri;
                    if (!Uri.TryCreate(value, UriKind.Absolute, out uri))
                    {
                        msg = "Not an absolute URI string.";
                        throw new ArgumentException(msg, value);
                    }

                    if (uri.Segments.Length > 1)
                    {
                        msg = "It includes the path segments.";
                        throw new ArgumentException(msg, value);
                    }
                }

                if (!canSet(out msg))
                {
                    Logger.Warning(msg);
                    return;
                }

                lock (_forState)
                {
                    if (!canSet(out msg))
                    {
                        Logger.Warning(msg);
                        return;
                    }

                    _origin = !value.IsNullOrEmpty() ? value.TrimEnd('/') : value;
                }
            }
        }

        public string Protocol
        {
            get
            {
                return _protocol ?? string.Empty;
            }

            internal set
            {
                _protocol = value;
            }
        }

        public WebSocketState ReadyState => _readyState;

        public SSLConfigurationClient SslConfiguration
        {
            get
            {
                if (!_client)
                {
                    var msg = "This instance is not a client.";
                    throw new InvalidOperationException(msg);
                }

                if (!IsSecure)
                {
                    var msg = "This instance does not use a secure connection.";
                    throw new InvalidOperationException(msg);
                }

                _sslConfig ??= new SSLConfigurationClient(_uri.DnsSafeHost);

                return _sslConfig;
            }
        }

        public Uri Url => _client ? _uri : _context.RequestUri;

        public TimeSpan WaitTime
        {
            get
            {
                return _waitTime;
            }

            set
            {
                if (value <= TimeSpan.Zero)
                {
                    throw new ArgumentOutOfRangeException(nameof(value), "Zero or less.");
                }

                string msg;
                if (!canSet(out msg))
                {
                    Logger.Warning(msg);
                    return;
                }

                lock (_forState)
                {
                    if (!canSet(out msg))
                    {
                        Logger.Warning(msg);
                        return;
                    }

                    _waitTime = value;
                }
            }
        }

        public event EventHandler<CloseEventArgs> OnDisconnect;

        public event EventHandler<ErrorEventArgs> OnError;

        public event EventHandler<MessageEventArgs> OnMessageReceived;

        public event EventHandler OnConnect;

        // As server
        private bool accept()
        {
            lock (_forState)
            {
                string msg;
                if (!checkIfAvailable(true, false, false, false, out msg))
                {
                    Logger.Error(msg);
                    error("An error has occurred in accepting.", null);

                    return false;
                }

                try
                {
                    if (!acceptHandshake())
                    {
                        return false;
                    }

                    _readyState = WebSocketState.Open;
                }
                catch (Exception ex)
                {
                    Logger.Error(ex.ToString());
                    fatal("An exception has occurred while accepting.", ex);

                    return false;
                }

                return true;
            }
        }

        // As server
        private bool acceptHandshake()
        {
            Logger.Debug(string.Format("A request from {0}:\n{1}", _context.UserEndPoint, _context));

            string msg;
            if (!checkHandshakeRequest(_context, out msg))
            {
                sendHttpResponse(createHandshakeFailureResponse(HttpStatusCode.BadRequest));

                Logger.Error(msg);
                fatal("An error has occurred while accepting.", CloseStatusCode.ProtocolError);

                return false;
            }

            if (!customCheckHandshakeRequest(_context, out msg))
            {
                sendHttpResponse(createHandshakeFailureResponse(HttpStatusCode.BadRequest));

                Logger.Error(msg);
                fatal("An error has occurred while accepting.", CloseStatusCode.PolicyViolation);

                return false;
            }

            _base64Key = _context.Headers["Sec-WebSocket-Key"];

            if (_protocol != null)
            {
                processSecWebSocketProtocolHeader(_context.SecWebSocketProtocols);
            }

            if (!IgnoreExtensions)
            {
                processSecWebSocketExtensionsClientHeader(_context.Headers["Sec-WebSocket-Extensions"]);
            }

            return sendHttpResponse(createHandshakeResponse());
        }

        private bool canSet(out string message)
        {
            message = null;

            if (_readyState == WebSocketState.Open)
            {
                message = "The connection has already been established.";
                return false;
            }

            if (_readyState == WebSocketState.Closing)
            {
                message = "The connection is closing.";
                return false;
            }

            return true;
        }

        // As server
        private bool checkHandshakeRequest(WebSocketContext context, out string message)
        {
            message = null;

            if (context.RequestUri == null)
            {
                message = "Specifies an invalid Request-URI.";
                return false;
            }

            if (!context.IsWebSocketRequest)
            {
                message = "Not a WebSocket handshake request.";
                return false;
            }

            var headers = context.Headers;
            if (!validateSecWebSocketKeyHeader(headers["Sec-WebSocket-Key"]))
            {
                message = "Includes no Sec-WebSocket-Key header, or it has an invalid value.";
                return false;
            }

            if (!validateSecWebSocketVersionClientHeader(headers["Sec-WebSocket-Version"]))
            {
                message = "Includes no Sec-WebSocket-Version header, or it has an invalid value.";
                return false;
            }

            if (!validateSecWebSocketProtocolClientHeader(headers["Sec-WebSocket-Protocol"]))
            {
                message = "Includes an invalid Sec-WebSocket-Protocol header.";
                return false;
            }

            if (!IgnoreExtensions
                && !validateSecWebSocketExtensionsClientHeader(headers["Sec-WebSocket-Extensions"])
            )
            {
                message = "Includes an invalid Sec-WebSocket-Extensions header.";
                return false;
            }

            return true;
        }

        // As client
        private bool checkHandshakeResponse(WebResponse response, out string message)
        {
            message = null;

            if (response.IsRedirect)
            {
                message = "Indicates the redirection.";
                return false;
            }

            if (response.IsUnauthorized)
            {
                message = "Requires the authentication.";
                return false;
            }

            if (!response.IsWebSocketResponse)
            {
                message = "Not a WebSocket handshake response.";
                return false;
            }

            var headers = response.Headers;
            if (!validateSecWebSocketAcceptHeader(headers["Sec-WebSocket-Accept"]))
            {
                message = "Includes no Sec-WebSocket-Accept header, or it has an invalid value.";
                return false;
            }

            if (!validateSecWebSocketProtocolServerHeader(headers["Sec-WebSocket-Protocol"]))
            {
                message = "Includes no Sec-WebSocket-Protocol header, or it has an invalid value.";
                return false;
            }

            if (!validateSecWebSocketExtensionsServerHeader(headers["Sec-WebSocket-Extensions"]))
            {
                message = "Includes an invalid Sec-WebSocket-Extensions header.";
                return false;
            }

            if (!validateSecWebSocketVersionServerHeader(headers["Sec-WebSocket-Version"]))
            {
                message = "Includes an invalid Sec-WebSocket-Version header.";
                return false;
            }

            return true;
        }

        private bool checkIfAvailable(
          bool connecting, bool open, bool closing, bool closed, out string message
        )
        {
            message = null;

            if (!connecting && _readyState == WebSocketState.Connecting)
            {
                message = "This operation is not available in: connecting";
                return false;
            }

            if (!open && _readyState == WebSocketState.Open)
            {
                message = "This operation is not available in: open";
                return false;
            }

            if (!closing && _readyState == WebSocketState.Closing)
            {
                message = "This operation is not available in: closing";
                return false;
            }

            if (!closed && _readyState == WebSocketState.Closed)
            {
                message = "This operation is not available in: closed";
                return false;
            }

            return true;
        }

        private bool checkIfAvailable(
          bool client,
          bool server,
          bool connecting,
          bool open,
          bool closing,
          bool closed,
          out string message
        )
        {
            message = null;

            if (!client && _client)
            {
                message = "This operation is not available in: client";
                return false;
            }

            if (!server && !_client)
            {
                message = "This operation is not available in: server";
                return false;
            }

            return checkIfAvailable(connecting, open, closing, closed, out message);
        }

        private static bool checkParametersForSetCredentials(
          string username, string password, out string message
        )
        {
            message = null;

            if (username.IsNullOrEmpty())
            {
                return true;
            }

            if (username.Contains(':') || !username.IsText())
            {
                message = "'username' contains an invalid character.";
                return false;
            }

            if (password.IsNullOrEmpty())
            {
                return true;
            }

            if (!password.IsText())
            {
                message = "'password' contains an invalid character.";
                return false;
            }

            return true;
        }

        private static bool checkParametersForSetProxy(
          string url, string username, string password, out string message
        )
        {
            message = null;

            if (url.IsNullOrEmpty())
            {
                return true;
            }

            Uri uri;
            if (!Uri.TryCreate(url, UriKind.Absolute, out uri)
                || uri.Scheme != "http"
                || uri.Segments.Length > 1
            )
            {
                message = "'url' is an invalid URL.";
                return false;
            }

            if (username.IsNullOrEmpty())
            {
                return true;
            }

            if (username.Contains(':') || !username.IsText())
            {
                message = "'username' contains an invalid character.";
                return false;
            }

            if (password.IsNullOrEmpty())
            {
                return true;
            }

            if (!password.IsText())
            {
                message = "'password' contains an invalid character.";
                return false;
            }

            return true;
        }

        private static bool checkProtocols(string[] protocols, out string message)
        {
            message = null;

            Func<string, bool> cond = protocol => protocol.IsNullOrEmpty()
                                                  || !protocol.IsToken();

            if (protocols.Contains(cond))
            {
                message = "It contains a value that is not a token.";
                return false;
            }

            if (protocols.ContainsTwice())
            {
                message = "It contains a value twice.";
                return false;
            }

            return true;
        }

        private bool checkReceivedFrame(WebSocketFrame frame, out string message)
        {
            message = null;

            var masked = frame.IsMasked;
            if (_client && masked)
            {
                message = "A frame from the server is masked.";
                return false;
            }

            if (!_client && !masked)
            {
                message = "A frame from a client is not masked.";
                return false;
            }

            if (_inContinuation && frame.IsData)
            {
                message = "A data frame has been received while receiving continuation frames.";
                return false;
            }

            if (frame.IsCompressed && _compression == CompressionMethod.None)
            {
                message = "A compressed frame has been received without any agreement for it.";
                return false;
            }

            if (frame.Rsv2 == Rsv.On)
            {
                message = "The RSV2 of a frame is non-zero without any negotiation for it.";
                return false;
            }

            if (frame.Rsv3 == Rsv.On)
            {
                message = "The RSV3 of a frame is non-zero without any negotiation for it.";
                return false;
            }

            return true;
        }

        private void close(ushort code, string reason)
        {
            if (_readyState == WebSocketState.Closing)
            {
                Logger.Info("The closing is already in progress.");
                return;
            }

            if (_readyState == WebSocketState.Closed)
            {
                Logger.Info("The connection has already been closed.");
                return;
            }

            if (code == 1005)
            { // == no status
                close(PayloadData.Empty, true, true, false);
                return;
            }

            var send = !code.IsReserved();
            close(new PayloadData(code, reason), send, send, false);
        }

        private void close(
          PayloadData payloadData, bool send, bool receive, bool received
        )
        {
            lock (_forState)
            {
                if (_readyState == WebSocketState.Closing)
                {
                    Logger.Info("The closing is already in progress.");
                    return;
                }

                if (_readyState == WebSocketState.Closed)
                {
                    Logger.Info("The connection has already been closed.");
                    return;
                }

                send = send && _readyState == WebSocketState.Open;
                receive = send && receive;

                _readyState = WebSocketState.Closing;
            }

            Logger.Trace("Begin closing the connection.");

            var res = closeHandshake(payloadData, send, receive, received);
            releaseResources();

            Logger.Trace("End closing the connection.");

            _readyState = WebSocketState.Closed;

            var e = new CloseEventArgs(payloadData);
            e.WasClean = res;

            try
            {
                OnDisconnect.Emit(this, e);
            }
            catch (Exception ex)
            {
                Logger.Error(ex.ToString());
                error("An error has occurred during the OnClose event.", ex);
            }
        }

        private void closeAsync(ushort code, string reason)
        {
            if (_readyState == WebSocketState.Closing)
            {
                Logger.Info("The closing is already in progress.");
                return;
            }

            if (_readyState == WebSocketState.Closed)
            {
                Logger.Info("The connection has already been closed.");
                return;
            }

            if (code == 1005)
            { // == no status
                closeAsync(PayloadData.Empty, true, true, false);
                return;
            }

            var send = !code.IsReserved();
            closeAsync(new PayloadData(code, reason), send, send, false);
        }

        private async Task closeAsync(PayloadData payloadData, bool send, bool receive, bool received)
        {
            await Task.Run(() => close(payloadData, send, receive, received));
        }

        private bool closeHandshake(byte[] frameAsBytes, bool receive, bool received)
        {
            var sent = frameAsBytes != null && sendBytes(frameAsBytes);

            var wait = !received && sent && receive && _receivingExited != null;
            if (wait)
            {
                received = _receivingExited.WaitOne(_waitTime);
            }

            var ret = sent && received;

            Logger.Debug(
              string.Format(
                "Was clean?: {0}\n  sent: {1}\n  received: {2}", ret, sent, received
              )
            );

            return ret;
        }

        private bool closeHandshake(
          PayloadData payloadData, bool send, bool receive, bool received
        )
        {
            var sent = false;
            if (send)
            {
                var frame = WebSocketFrame.CreateCloseFrame(payloadData, _client);
                sent = sendBytes(frame.ToArray());

                if (_client)
                {
                    frame.Unmask();
                }
            }

            var wait = !received && sent && receive && _receivingExited != null;
            if (wait)
            {
                received = _receivingExited.WaitOne(_waitTime);
            }

            var ret = sent && received;

            Logger.Debug(
              string.Format(
                "Was clean?: {0}\n  sent: {1}\n  received: {2}", ret, sent, received
              )
            );

            return ret;
        }

        // As client
        private bool connect()
        {
            lock (_forState)
            {
                string msg;
                if (!checkIfAvailable(true, false, false, true, out msg))
                {
                    Logger.Error(msg);
                    error("An error has occurred in connecting.", null);

                    return false;
                }

                if (_retryCountForConnect > _maxRetryCountForConnect)
                {
                    _retryCountForConnect = 0;
                    Logger.Error("A series of reconnecting has failed.");

                    return false;
                }

                _readyState = WebSocketState.Connecting;

                try
                {
                    doHandshake();
                }
                catch (Exception ex)
                {
                    _retryCountForConnect++;
                    Logger.Error(ex.ToString());
                    fatal("An exception has occurred while connecting.", ex);

                    return false;
                }

                _retryCountForConnect = 1;
                _readyState = WebSocketState.Open;

                return true;
            }
        }

        // As client
        private string createExtensions()
        {
            var buff = new StringBuilder(80);

            if (_compression != CompressionMethod.None)
            {
                var str = _compression.ToExtensionString(
                  "server_no_context_takeover", "client_no_context_takeover");

                buff.AppendFormat("{0}, ", str);
            }

            var len = buff.Length;
            if (len > 2)
            {
                buff.Length = len - 2;
                return buff.ToString();
            }

            return null;
        }

        // As server
        private WebResponse createHandshakeFailureResponse(HttpStatusCode code)
        {
            var ret = WebResponse.CreateCloseResponse(code);
            ret.Headers["Sec-WebSocket-Version"] = _version;

            return ret;
        }

        // As client
        private WebRequest createHandshakeRequest()
        {
            var ret = WebRequest.CreateWebSocketRequest(_uri);

            var headers = ret.Headers;
            if (!_origin.IsNullOrEmpty())
            {
                headers["Origin"] = _origin;
            }

            headers["Sec-WebSocket-Key"] = _base64Key;

            _protocolsRequested = _protocols != null;
            if (_protocolsRequested)
            {
                headers["Sec-WebSocket-Protocol"] = _protocols.ToString(", ");
            }

            _extensionsRequested = _compression != CompressionMethod.None;
            if (_extensionsRequested)
            {
                headers["Sec-WebSocket-Extensions"] = createExtensions();
            }

            headers["Sec-WebSocket-Version"] = _version;

            AuthenticationResponse authRes = null;
            if (_authChallenge != null && Credentials != null)
            {
                authRes = new AuthenticationResponse(_authChallenge, Credentials, _nonceCount);
                _nonceCount = authRes.NonceCount;
            }
            else if (_preAuth)
            {
                authRes = new AuthenticationResponse(Credentials);
            }

            if (authRes != null)
            {
                headers["Authorization"] = authRes.ToString();
            }

            if (CookieCollection.Count > 0)
            {
                ret.SetCookies(CookieCollection);
            }

            return ret;
        }

        // As server
        private WebResponse createHandshakeResponse()
        {
            var ret = WebResponse.CreateWebSocketResponse();

            var headers = ret.Headers;
            headers["Sec-WebSocket-Accept"] = CreateResponseKey(_base64Key);

            if (_protocol != null)
            {
                headers["Sec-WebSocket-Protocol"] = _protocol;
            }

            if (_extensions != null)
            {
                headers["Sec-WebSocket-Extensions"] = _extensions;
            }

            if (CookieCollection.Count > 0)
            {
                ret.SetCookies(CookieCollection);
            }

            return ret;
        }

        // As server
        private bool customCheckHandshakeRequest(WebSocketContext context, out string message)
        {
            message = null;
            return CustomHandshakeRequestChecker == null
                   || (message = CustomHandshakeRequestChecker(context)) == null;
        }

        private MessageEventArgs dequeueFromMessageEventQueue()
        {
            lock (_forMessageEventQueue)
            {
                return _messageEventQueue.Count > 0 ? _messageEventQueue.Dequeue() : null;
            }
        }

        // As client
        private void doHandshake()
        {
            setClientStream();
            var res = sendHandshakeRequest();

            string msg;
            if (!checkHandshakeResponse(res, out msg))
            {
                throw new WebSocketException(CloseStatusCode.ProtocolError, msg);
            }

            if (_protocolsRequested)
            {
                _protocol = res.Headers["Sec-WebSocket-Protocol"];
            }

            if (_extensionsRequested)
            {
                processSecWebSocketExtensionsServerHeader(res.Headers["Sec-WebSocket-Extensions"]);
            }

            processCookies(res.Cookies);
        }

        private void enqueueToMessageEventQueue(MessageEventArgs e)
        {
            lock (_forMessageEventQueue)
            {
                _messageEventQueue.Enqueue(e);
            }
        }

        private void error(string message, Exception exception)
        {
            try
            {
                OnError.Emit(this, new ErrorEventArgs(message, exception));
            }
            catch (Exception ex)
            {
                Logger.Error(ex.ToString());
            }
        }

        private void fatal(string message, Exception exception)
        {
            var code = exception is WebSocketException
                       ? ((WebSocketException)exception).Code
                       : CloseStatusCode.Abnormal;

            fatal(message, (ushort)code);
        }

        private void fatal(string message, ushort code)
        {
            var payload = new PayloadData(code, message);
            close(payload, !code.IsReserved(), false, false);
        }

        private void fatal(string message, CloseStatusCode code)
        {
            fatal(message, (ushort)code);
        }

        private void init()
        {
            _compression = CompressionMethod.None;
            CookieCollection = new CookieCollection();
            _forPing = new object();
            _forSend = new object();
            _forState = new object();
            _messageEventQueue = new Queue<MessageEventArgs>();
            _forMessageEventQueue = ((ICollection)_messageEventQueue).SyncRoot;
            _readyState = WebSocketState.Connecting;
        }

        private void message()
        {
            MessageEventArgs e = null;
            lock (_forMessageEventQueue)
            {
                if (_inMessage || _messageEventQueue.Count == 0 || _readyState != WebSocketState.Open)
                {
                    return;
                }

                _inMessage = true;
                e = _messageEventQueue.Dequeue();
            }

            _message(e);
        }

        private void messagec(MessageEventArgs e)
        {
            do
            {
                try
                {
                    OnMessageReceived.Emit(this, e);
                }
                catch (Exception ex)
                {
                    Logger.Error(ex.ToString());
                    error("An error has occurred during an OnMessage event.", ex);
                }

                lock (_forMessageEventQueue)
                {
                    if (_messageEventQueue.Count == 0 || _readyState != WebSocketState.Open)
                    {
                        _inMessage = false;
                        break;
                    }

                    e = _messageEventQueue.Dequeue();
                }
            }
            while (true);
        }

        private void messages(MessageEventArgs e)
        {
            try
            {
                OnMessageReceived.Emit(this, e);
            }
            catch (Exception ex)
            {
                Logger.Error(ex.ToString());
                error("An error has occurred during an OnMessage event.", ex);
            }

            lock (_forMessageEventQueue)
            {
                if (_messageEventQueue.Count == 0 || _readyState != WebSocketState.Open)
                {
                    _inMessage = false;
                    return;
                }

                e = _messageEventQueue.Dequeue();
            }

            ThreadPool.QueueUserWorkItem(state => messages(e));
        }

        private void open()
        {
            _inMessage = true;
            startReceiving();
            try
            {
                OnConnect.Emit(this, EventArgs.Empty);
            }
            catch (Exception ex)
            {
                Logger.Error(ex.ToString());
                error("An error has occurred during the OnOpen event.", ex);
            }

            MessageEventArgs e = null;
            lock (_forMessageEventQueue)
            {
                if (_messageEventQueue.Count == 0 || _readyState != WebSocketState.Open)
                {
                    _inMessage = false;
                    return;
                }

                e = _messageEventQueue.Dequeue();
            }

            Task.Run(() => _message(e));
        }

        private bool ping(byte[] data)
        {
            if (_readyState != WebSocketState.Open)
            {
                return false;
            }

            var pongReceived = _pongReceived;
            if (pongReceived == null)
            {
                return false;
            }

            lock (_forPing)
            {
                try
                {
                    pongReceived.Reset();
                    if (!send(FinalFrame.Final, Opcode.Ping, data, false))
                    {
                        return false;
                    }

                    return pongReceived.WaitOne(_waitTime);
                }
                catch (ObjectDisposedException)
                {
                    return false;
                }
            }
        }

        private bool processCloseFrame(WebSocketFrame frame)
        {
            var payload = frame.PayloadData;
            close(payload, !payload.HasReservedCode, false, true);

            return false;
        }

        // As client
        private void processCookies(CookieCollection cookies)
        {
            if (cookies.Count == 0)
            {
                return;
            }

            CookieCollection.SetOrRemove(cookies);
        }

        private bool processDataFrame(WebSocketFrame frame)
        {
            enqueueToMessageEventQueue(
              frame.IsCompressed
              ? new MessageEventArgs(
                  frame.Opcode, frame.PayloadData.ApplicationData.Decompress(_compression))
              : new MessageEventArgs(frame));

            return true;
        }

        private bool processFragmentFrame(WebSocketFrame frame)
        {
            if (!_inContinuation)
            {
                // Must process first fragment.
                if (frame.IsContinuation)
                {
                    return true;
                }

                _fragmentsOpcode = frame.Opcode;
                _fragmentsCompressed = frame.IsCompressed;
                _fragmentsBuffer = new MemoryStream();
                _inContinuation = true;
            }

            _fragmentsBuffer.WriteBytes(frame.PayloadData.ApplicationData, 1024);
            if (frame.IsFinal)
            {
                using (_fragmentsBuffer)
                {
                    var data = _fragmentsCompressed
                               ? _fragmentsBuffer.DecompressToArray(_compression)
                               : _fragmentsBuffer.ToArray();

                    enqueueToMessageEventQueue(new MessageEventArgs(_fragmentsOpcode, data));
                }

                _fragmentsBuffer = null;
                _inContinuation = false;
            }

            return true;
        }

        private bool processPingFrame(WebSocketFrame frame)
        {
            Logger.Trace("A ping was received.");

            var pong = WebSocketFrame.CreatePongFrame(frame.PayloadData, _client);

            lock (_forState)
            {
                if (_readyState != WebSocketState.Open)
                {
                    Logger.Error("The connection is closing.");
                    return true;
                }

                if (!sendBytes(pong.ToArray()))
                {
                    return false;
                }
            }

            Logger.Trace("A pong to this ping has been sent.");

            if (CallMessageOnPing)
            {
                if (_client)
                {
                    pong.Unmask();
                }

                enqueueToMessageEventQueue(new MessageEventArgs(frame));
            }

            return true;
        }

        private bool processPongFrame(WebSocketFrame frame)
        {
            Logger.Trace("A pong was received.");

            try
            {
                _pongReceived.Set();
            }
            catch (NullReferenceException ex)
            {
                Logger.Error(ex.Message);
                Logger.Debug(ex.ToString());

                return false;
            }
            catch (ObjectDisposedException ex)
            {
                Logger.Error(ex.Message);
                Logger.Debug(ex.ToString());

                return false;
            }

            Logger.Trace("It has been signaled.");

            return true;
        }

        private bool processReceivedFrame(WebSocketFrame frame)
        {
            string msg;
            if (!checkReceivedFrame(frame, out msg))
            {
                throw new WebSocketException(CloseStatusCode.ProtocolError, msg);
            }

            frame.Unmask();
            return frame.IsFragment
                   ? processFragmentFrame(frame)
                   : frame.IsData
                     ? processDataFrame(frame)
                     : frame.IsPing
                       ? processPingFrame(frame)
                       : frame.IsPong
                         ? processPongFrame(frame)
                         : frame.IsClose
                           ? processCloseFrame(frame)
                           : processUnsupportedFrame(frame);
        }

        // As server
        private void processSecWebSocketExtensionsClientHeader(string value)
        {
            if (value == null)
            {
                return;
            }

            var buff = new StringBuilder(80);

            var comp = false;
            foreach (var e in value.SplitHeaderValue(','))
            {
                var ext = e.Trim();
                if (!comp && ext.IsCompressionExtension(CompressionMethod.Deflate))
                {
                    _compression = CompressionMethod.Deflate;
                    buff.AppendFormat(
                      "{0}, ",
                      _compression.ToExtensionString(
                        "client_no_context_takeover", "server_no_context_takeover"
                      )
                    );

                    comp = true;
                }
            }

            var len = buff.Length;
            if (len > 2)
            {
                buff.Length = len - 2;
                _extensions = buff.ToString();
            }
        }

        // As client
        private void processSecWebSocketExtensionsServerHeader(string value)
        {
            if (value == null)
            {
                _compression = CompressionMethod.None;
                return;
            }

            _extensions = value;
        }

        // As server
        private void processSecWebSocketProtocolHeader(IEnumerable<string> values)
        {
            if (values.Contains(p => p == _protocol))
            {
                return;
            }

            _protocol = null;
        }

        private bool processUnsupportedFrame(WebSocketFrame frame)
        {
            Logger.Error("An unsupported frame:" + frame.PrintToString(false));
            fatal("There is no way to handle it.", CloseStatusCode.PolicyViolation);

            return false;
        }

        // As client
        private void releaseClientResources()
        {
            if (_stream != null)
            {
                _stream.Dispose();
                _stream = null;
            }

            if (_tcpClient != null)
            {
                _tcpClient.Close();
                _tcpClient = null;
            }
        }

        private void releaseCommonResources()
        {
            if (_fragmentsBuffer != null)
            {
                _fragmentsBuffer.Dispose();
                _fragmentsBuffer = null;
                _inContinuation = false;
            }

            if (_pongReceived != null)
            {
                _pongReceived.Close();
                _pongReceived = null;
            }

            if (_receivingExited != null)
            {
                _receivingExited.Close();
                _receivingExited = null;
            }
        }

        private void releaseResources()
        {
            if (_client)
            {
                releaseClientResources();
            }
            else
            {
                releaseServerResources();
            }

            releaseCommonResources();
        }

        // As server
        private void releaseServerResources()
        {
            if (_closeContext == null)
            {
                return;
            }

            _closeContext();
            _closeContext = null;
            _stream = null;
            _context = null;
        }

        private bool send(Opcode opcode, Stream stream)
        {
            lock (_forSend)
            {
                var src = stream;
                var compressed = false;
                var sent = false;
                try
                {
                    if (_compression != CompressionMethod.None)
                    {
                        stream = stream.Compress(_compression);
                        compressed = true;
                    }

                    sent = send(opcode, stream, compressed);
                    if (!sent)
                    {
                        error("A send has been interrupted.", null);
                    }
                }
                catch (Exception ex)
                {
                    Logger.Error(ex.ToString());
                    error("An error has occurred during a send.", ex);
                }
                finally
                {
                    if (compressed)
                    {
                        stream.Dispose();
                    }

                    src.Dispose();
                }

                return sent;
            }
        }

        private bool send(Opcode opcode, Stream stream, bool compressed)
        {
            var len = stream.Length;
            if (len == 0)
            {
                return send(FinalFrame.Final, opcode, EmptyBytes, false);
            }

            var quo = len / FragmentLength;
            var rem = (int)(len % FragmentLength);

            byte[] buff = null;
            if (quo == 0)
            {
                buff = new byte[rem];
                return stream.Read(buff, 0, rem) == rem
                       && send(FinalFrame.Final, opcode, buff, compressed);
            }

            if (quo == 1 && rem == 0)
            {
                buff = new byte[FragmentLength];
                return stream.Read(buff, 0, FragmentLength) == FragmentLength
                       && send(FinalFrame.Final, opcode, buff, compressed);
            }

            /* Send fragments */

            // Begin
            buff = new byte[FragmentLength];
            var sent = stream.Read(buff, 0, FragmentLength) == FragmentLength
                       && send(FinalFrame.More, opcode, buff, compressed);

            if (!sent)
            {
                return false;
            }

            var n = rem == 0 ? quo - 2 : quo - 1;
            for (long i = 0; i < n; i++)
            {
                sent = stream.Read(buff, 0, FragmentLength) == FragmentLength
                       && send(FinalFrame.More, Opcode.Cont, buff, false);

                if (!sent)
                {
                    return false;
                }
            }

            // End
            if (rem == 0)
            {
                rem = FragmentLength;
            }
            else
            {
                buff = new byte[rem];
            }

            return stream.Read(buff, 0, rem) == rem
                   && send(FinalFrame.Final, Opcode.Cont, buff, false);
        }

        private bool send(FinalFrame fin, Opcode opcode, byte[] data, bool compressed)
        {
            lock (_forState)
            {
                if (_readyState != WebSocketState.Open)
                {
                    Logger.Error("The connection is closing.");
                    return false;
                }

                var frame = new WebSocketFrame(fin, opcode, data, compressed, _client);
                return sendBytes(frame.ToArray());
            }
        }

        private void sendAsync(Opcode opcode, Stream stream, Action<bool> completed)
        {
            Func<Opcode, Stream, bool> sender = send;
            Task.Run(() =>
            {
                try
                {
                    var sent = sender(opcode, stream);
                    completed?.Invoke(sent);
                }
                catch (Exception ex)
                {
                    Logger.Error(ex.ToString());
                    error("An error has occurred during the callback for an async send.", ex);
                }
            });
        }

        private bool sendBytes(byte[] bytes)
        {
            try
            {
                _stream.Write(bytes, 0, bytes.Length);
            }
            catch (Exception ex)
            {
                Logger.Error(ex.Message);
                Logger.Debug(ex.ToString());

                return false;
            }

            return true;
        }

        // As client
        private WebResponse sendHandshakeRequest()
        {
            var req = createHandshakeRequest();
            var res = sendHttpRequest(req, 90000);
            if (res.IsUnauthorized)
            {
                var chal = res.Headers["WWW-Authenticate"];
                Logger.Warning(string.Format("Received an authentication requirement for '{0}'.", chal));
                if (chal.IsNullOrEmpty())
                {
                    Logger.Error("No authentication challenge is specified.");
                    return res;
                }

                _authChallenge = AuthenticationChallenge.Parse(chal);
                if (_authChallenge == null)
                {
                    Logger.Error("An invalid authentication challenge is specified.");
                    return res;
                }

                if (Credentials != null &&
                    (!_preAuth || _authChallenge.Scheme == AuthenticationSchemes.Digest))
                {
                    if (res.HasConnectionClose)
                    {
                        releaseClientResources();
                        setClientStream();
                    }

                    var authRes = new AuthenticationResponse(_authChallenge, Credentials, _nonceCount);
                    _nonceCount = authRes.NonceCount;
                    req.Headers["Authorization"] = authRes.ToString();
                    res = sendHttpRequest(req, 15000);
                }
            }

            if (res.IsRedirect)
            {
                var url = res.Headers["Location"];
                Logger.Warning(string.Format("Received a redirection to '{0}'.", url));
                if (_isRedirectionEnabled)
                {
                    if (url.IsNullOrEmpty())
                    {
                        Logger.Error("No url to redirect is located.");
                        return res;
                    }

                    Uri uri;
                    string msg;
                    if (!url.TryCreateWebSocketUri(out uri, out msg))
                    {
                        Logger.Error("An invalid url to redirect is located: " + msg);
                        return res;
                    }

                    releaseClientResources();

                    _uri = uri;
                    IsSecure = uri.Scheme == "wss";

                    setClientStream();
                    return sendHandshakeRequest();
                }
            }

            return res;
        }

        // As client
        private WebResponse sendHttpRequest(WebRequest request, int millisecondsTimeout)
        {
            Logger.Debug("Request to server:\n" + request.ToString());
            var res = request.GetResponse(_stream, millisecondsTimeout);
            Logger.Debug("Response to request:\n" + res.ToString());

            return res;
        }

        // As server
        private bool sendHttpResponse(WebResponse response)
        {
            Logger.Debug("Response to request:\n" + response.ToString());
            return sendBytes(response.ToByteArray());
        }

        // As client
        private void sendProxyConnectRequest()
        {
            var req = WebRequest.CreateConnectRequest(_uri);
            var res = sendHttpRequest(req, 90000);
            if (res.IsProxyAuthenticationRequired)
            {
                var chal = res.Headers["Proxy-Authenticate"];
                Logger.Warning(
                  string.Format("Received a proxy authentication requirement for '{0}'.", chal));

                if (chal.IsNullOrEmpty())
                {
                    throw new WebSocketException("No proxy authentication challenge is specified.");
                }

                var authChal = AuthenticationChallenge.Parse(chal);
                if (authChal == null)
                {
                    throw new WebSocketException("An invalid proxy authentication challenge is specified.");
                }

                if (_proxyCredentials != null)
                {
                    if (res.HasConnectionClose)
                    {
                        releaseClientResources();
                        _tcpClient = new TcpClient(_proxyUri.DnsSafeHost, _proxyUri.Port);
                        _stream = _tcpClient.GetStream();
                    }

                    var authRes = new AuthenticationResponse(authChal, _proxyCredentials, 0);
                    req.Headers["Proxy-Authorization"] = authRes.ToString();
                    res = sendHttpRequest(req, 15000);
                }

                if (res.IsProxyAuthenticationRequired)
                {
                    throw new WebSocketException("A proxy authentication is required.");
                }
            }

            if (res.StatusCode[0] != '2')
            {
                throw new WebSocketException(
                  "The proxy has failed a connection to the requested host and port.");
            }
        }

        // As client
        private void setClientStream()
        {
            if (_proxyUri != null)
            {
                _tcpClient = new TcpClient(_proxyUri.DnsSafeHost, _proxyUri.Port);
                _stream = _tcpClient.GetStream();
                sendProxyConnectRequest();
            }
            else
            {
                _tcpClient = new TcpClient(_uri.DnsSafeHost, _uri.Port);
                _stream = _tcpClient.GetStream();
            }

            if (IsSecure)
            {
                var conf = SslConfiguration;
                var host = conf.TargetHost;
                if (host != _uri.DnsSafeHost)
                {
                    throw new WebSocketException(
                      CloseStatusCode.TlsHandshakeFailure, "An invalid host name is specified.");
                }

                try
                {
                    var sslStream = new SslStream(
                      _stream,
                      false,
                      conf.ServerCertificateValidationCallback,
                      conf.ClientCertificateSelectionCallback);

                    sslStream.AuthenticateAsClient(
                      host,
                      conf.Certificates,
                      conf.SslProtocols,
                      conf.CheckForCertificateRevocation);

                    _stream = sslStream;
                }
                catch (Exception ex)
                {
                    throw new WebSocketException(CloseStatusCode.TlsHandshakeFailure, ex);
                }
            }
        }

        private void startReceiving()
        {
            if (_messageEventQueue.Count > 0)
            {
                _messageEventQueue.Clear();
            }

            _pongReceived = new ManualResetEvent(false);
            _receivingExited = new ManualResetEvent(false);

            Action receive = null;
            receive =
              () =>
                WebSocketFrame.ReadFrameAsync(
                  _stream,
                  false,
                  frame =>
                  {
                      if (!processReceivedFrame(frame) || _readyState == WebSocketState.Closed)
                      {
                          var exited = _receivingExited;
                          if (exited != null)
                          {
                              exited.Set();
                          }

                          return;
                      }

                      // Receive next asap because the Ping or Close needs a response to it.
                      receive();

                      if (_inMessage || !HasMessage || _readyState != WebSocketState.Open)
                      {
                          return;
                      }

                      message();
                  },
                  ex =>
                  {
                      Logger.Error(ex.ToString());
                      fatal("An exception has occurred while receiving.", ex);
                  }
                );

            receive();
        }

        // As client
        private bool validateSecWebSocketAcceptHeader(string value)
        {
            return value != null && value == CreateResponseKey(_base64Key);
        }

        // As server
        private bool validateSecWebSocketExtensionsClientHeader(string value)
        {
            return value == null || value.Length > 0;
        }

        // As client
        private bool validateSecWebSocketExtensionsServerHeader(string value)
        {
            if (value == null)
            {
                return true;
            }

            if (value.Length == 0)
            {
                return false;
            }

            if (!_extensionsRequested)
            {
                return false;
            }

            var comp = _compression != CompressionMethod.None;
            foreach (var e in value.SplitHeaderValue(','))
            {
                var ext = e.Trim();
                if (comp && ext.IsCompressionExtension(_compression))
                {
                    if (!ext.Contains("server_no_context_takeover"))
                    {
                        Logger.Error("The server hasn't sent back 'server_no_context_takeover'.");
                        return false;
                    }

                    if (!ext.Contains("client_no_context_takeover"))
                    {
                        Logger.Warning("The server hasn't sent back 'client_no_context_takeover'.");
                    }

                    var method = _compression.ToExtensionString();
                    var invalid =
                      ext.SplitHeaderValue(';').Contains(
                        t =>
                        {
                            t = t.Trim();
                            return t != method
                         && t != "server_no_context_takeover"
                         && t != "client_no_context_takeover";
                        }
                      );

                    if (invalid)
                    {
                        return false;
                    }
                }
                else
                {
                    return false;
                }
            }

            return true;
        }

        // As server
        private bool validateSecWebSocketKeyHeader(string value)
        {
            return value != null && value.Length > 0;
        }

        // As server
        private bool validateSecWebSocketProtocolClientHeader(string value)
        {
            return value == null || value.Length > 0;
        }

        // As client
        private bool validateSecWebSocketProtocolServerHeader(string value)
        {
            if (value == null)
            {
                return !_protocolsRequested;
            }

            if (value.Length == 0)
            {
                return false;
            }

            return _protocolsRequested && _protocols.Contains(p => p == value);
        }

        // As server
        private bool validateSecWebSocketVersionClientHeader(string value)
        {
            return value != null && value == _version;
        }

        // As client
        private bool validateSecWebSocketVersionServerHeader(string value)
        {
            return value == null || value == _version;
        }

        // As server
        internal void Close(WebResponse response)
        {
            _readyState = WebSocketState.Closing;

            sendHttpResponse(response);
            releaseServerResources();

            _readyState = WebSocketState.Closed;
        }

        // As server
        internal void Close(HttpStatusCode code)
        {
            Close(createHandshakeFailureResponse(code));
        }

        // As server
        internal void Close(PayloadData payloadData, byte[] frameAsBytes)
        {
            lock (_forState)
            {
                if (_readyState == WebSocketState.Closing)
                {
                    Logger.Info("The closing is already in progress.");
                    return;
                }

                if (_readyState == WebSocketState.Closed)
                {
                    Logger.Info("The connection has already been closed.");
                    return;
                }

                _readyState = WebSocketState.Closing;
            }

            Logger.Trace("Begin closing the connection.");

            var sent = frameAsBytes != null && sendBytes(frameAsBytes);
            var received = sent && _receivingExited != null
&& _receivingExited.WaitOne(_waitTime);

            var res = sent && received;

            Logger.Debug(
              string.Format(
                "Was clean?: {0}\n  sent: {1}\n  received: {2}", res, sent, received
              )
            );

            releaseServerResources();
            releaseCommonResources();

            Logger.Trace("End closing the connection.");

            _readyState = WebSocketState.Closed;

            var e = new CloseEventArgs(payloadData);
            e.WasClean = res;

            try
            {
                OnDisconnect.Emit(this, e);
            }
            catch (Exception ex)
            {
                Logger.Error(ex.ToString());
            }
        }

        // As client
        internal static string CreateBase64Key()
        {
            var src = new byte[16];
            RandomNumber.GetBytes(src);

            return Convert.ToBase64String(src);
        }

        internal static string CreateResponseKey(string base64Key)
        {
            var buff = new StringBuilder(base64Key, 64);
            buff.Append(_guid);
            SHA1 sha1 = new SHA1CryptoServiceProvider();
            var src = sha1.ComputeHash(buff.ToString().UTF8Encode());

            return Convert.ToBase64String(src);
        }

        // As server
        internal void InternalAccept()
        {
            try
            {
                if (!acceptHandshake())
                {
                    return;
                }

                _readyState = WebSocketState.Open;
            }
            catch (Exception ex)
            {
                Logger.Error(ex.ToString());
                fatal("An exception has occurred while accepting.", ex);

                return;
            }

            open();
        }

        // As server
        internal bool Ping(byte[] frameAsBytes, TimeSpan timeout)
        {
            if (_readyState != WebSocketState.Open)
            {
                return false;
            }

            var pongReceived = _pongReceived;
            if (pongReceived == null)
            {
                return false;
            }

            lock (_forPing)
            {
                try
                {
                    pongReceived.Reset();

                    lock (_forState)
                    {
                        if (_readyState != WebSocketState.Open)
                        {
                            return false;
                        }

                        if (!sendBytes(frameAsBytes))
                        {
                            return false;
                        }
                    }

                    return pongReceived.WaitOne(timeout);
                }
                catch (ObjectDisposedException)
                {
                    return false;
                }
            }
        }

        // As server
        internal void Send(
          Opcode opcode, byte[] data, Dictionary<CompressionMethod, byte[]> cache
        )
        {
            lock (_forSend)
            {
                lock (_forState)
                {
                    if (_readyState != WebSocketState.Open)
                    {
                        Logger.Error("The connection is closing.");
                        return;
                    }

                    byte[] found;
                    if (!cache.TryGetValue(_compression, out found))
                    {
                        found = new WebSocketFrame(
                                  FinalFrame.Final,
                                  opcode,
                                  data.Compress(_compression),
                                  _compression != CompressionMethod.None,
                                  false
                                )
                                .ToArray();

                        cache.Add(_compression, found);
                    }

                    sendBytes(found);
                }
            }
        }

        // As server
        internal void Send(
          Opcode opcode, Stream stream, Dictionary<CompressionMethod, Stream> cache
        )
        {
            lock (_forSend)
            {
                Stream found;
                if (!cache.TryGetValue(_compression, out found))
                {
                    found = stream.Compress(_compression);
                    cache.Add(_compression, found);
                }
                else
                {
                    found.Position = 0;
                }

                send(opcode, found, _compression != CompressionMethod.None);
            }
        }

        public void Accept()
        {
            string msg;
            if (!checkIfAvailable(false, true, true, false, false, false, out msg))
            {
                Logger.Error(msg);
                error("An error has occurred in accepting.", null);

                return;
            }

            if (accept())
            {
                open();
            }
        }

        public void AcceptAsync()
        {
            string msg;
            if (!checkIfAvailable(false, true, true, false, false, false, out msg))
            {
                Logger.Error(msg);
                error("An error has occurred in accepting.", null);
                return;
            }

            Func<bool> acceptor = accept;
            Task.Run(() =>
            {
                if (acceptor())
                {
                    open();
                }
            });
        }

        public void Close()
        {
            close(1005, string.Empty);
        }

        public void Close(ushort code)
        {
            if (!code.IsCloseStatusCode())
            {
                var msg = "Less than 1000 or greater than 4999.";
                throw new ArgumentOutOfRangeException(nameof(code), msg);
            }

            if (_client && code == 1011)
            {
                var msg = "1011 cannot be used.";
                throw new ArgumentException(msg, nameof(code));
            }

            if (!_client && code == 1010)
            {
                var msg = "1010 cannot be used.";
                throw new ArgumentException(msg, nameof(code));
            }

            close(code, string.Empty);
        }

        public void Close(CloseStatusCode code)
        {
            if (_client && code == CloseStatusCode.ServerError)
            {
                var msg = "ServerError cannot be used.";
                throw new ArgumentException(msg, nameof(code));
            }

            if (!_client && code == CloseStatusCode.MandatoryExtension)
            {
                var msg = "MandatoryExtension cannot be used.";
                throw new ArgumentException(msg, nameof(code));
            }

            close((ushort)code, string.Empty);
        }

        public void Close(ushort code, string reason)
        {
            if (!code.IsCloseStatusCode())
            {
                var msg = "Less than 1000 or greater than 4999.";
                throw new ArgumentOutOfRangeException(nameof(code), msg);
            }

            if (_client && code == 1011)
            {
                var msg = "1011 cannot be used.";
                throw new ArgumentException(msg, nameof(code));
            }

            if (!_client && code == 1010)
            {
                var msg = "1010 cannot be used.";
                throw new ArgumentException(msg, nameof(code));
            }

            if (reason.IsNullOrEmpty())
            {
                close(code, string.Empty);
                return;
            }

            if (code == 1005)
            {
                var msg = "1005 cannot be used.";
                throw new ArgumentException(msg, nameof(code));
            }

            byte[] bytes;
            if (!reason.TryGetUTF8EncodedBytes(out bytes))
            {
                var msg = "It could not be UTF-8-encoded.";
                throw new ArgumentException(msg, nameof(reason));
            }

            if (bytes.Length > 123)
            {
                var msg = "Its size is greater than 123 bytes.";
                throw new ArgumentOutOfRangeException(nameof(reason), msg);
            }

            close(code, reason);
        }

        public void Close(CloseStatusCode code, string reason)
        {
            if (_client && code == CloseStatusCode.ServerError)
            {
                var msg = "ServerError cannot be used.";
                throw new ArgumentException(msg, nameof(code));
            }

            if (!_client && code == CloseStatusCode.MandatoryExtension)
            {
                var msg = "MandatoryExtension cannot be used.";
                throw new ArgumentException(msg, nameof(code));
            }

            if (reason.IsNullOrEmpty())
            {
                close((ushort)code, string.Empty);
                return;
            }

            if (code == CloseStatusCode.NoStatus)
            {
                var msg = "NoStatus cannot be used.";
                throw new ArgumentException(msg, nameof(code));
            }

            byte[] bytes;
            if (!reason.TryGetUTF8EncodedBytes(out bytes))
            {
                var msg = "It could not be UTF-8-encoded.";
                throw new ArgumentException(msg, nameof(reason));
            }

            if (bytes.Length > 123)
            {
                var msg = "Its size is greater than 123 bytes.";
                throw new ArgumentOutOfRangeException(nameof(reason), msg);
            }

            close((ushort)code, reason);
        }

        public void CloseAsync()
        {
            closeAsync(1005, string.Empty);
        }

        public void CloseAsync(ushort code)
        {
            if (!code.IsCloseStatusCode())
            {
                var msg = "Less than 1000 or greater than 4999.";
                throw new ArgumentOutOfRangeException(nameof(code), msg);
            }

            if (_client && code == 1011)
            {
                var msg = "1011 cannot be used.";
                throw new ArgumentException(msg, nameof(code));
            }

            if (!_client && code == 1010)
            {
                var msg = "1010 cannot be used.";
                throw new ArgumentException(msg, nameof(code));
            }

            closeAsync(code, string.Empty);
        }

        public void CloseAsync(CloseStatusCode code)
        {
            if (_client && code == CloseStatusCode.ServerError)
            {
                var msg = "ServerError cannot be used.";
                throw new ArgumentException(msg, nameof(code));
            }

            if (!_client && code == CloseStatusCode.MandatoryExtension)
            {
                var msg = "MandatoryExtension cannot be used.";
                throw new ArgumentException(msg, nameof(code));
            }

            closeAsync((ushort)code, string.Empty);
        }

        public void CloseAsync(ushort code, string reason)
        {
            if (!code.IsCloseStatusCode())
            {
                var msg = "Less than 1000 or greater than 4999.";
                throw new ArgumentOutOfRangeException(nameof(code), msg);
            }

            if (_client && code == 1011)
            {
                var msg = "1011 cannot be used.";
                throw new ArgumentException(msg, nameof(code));
            }

            if (!_client && code == 1010)
            {
                var msg = "1010 cannot be used.";
                throw new ArgumentException(msg, nameof(code));
            }

            if (reason.IsNullOrEmpty())
            {
                closeAsync(code, string.Empty);
                return;
            }

            if (code == 1005)
            {
                var msg = "1005 cannot be used.";
                throw new ArgumentException(msg, nameof(code));
            }

            byte[] bytes;
            if (!reason.TryGetUTF8EncodedBytes(out bytes))
            {
                var msg = "It could not be UTF-8-encoded.";
                throw new ArgumentException(msg, nameof(reason));
            }

            if (bytes.Length > 123)
            {
                var msg = "Its size is greater than 123 bytes.";
                throw new ArgumentOutOfRangeException(nameof(reason), msg);
            }

            closeAsync(code, reason);
        }

        public void CloseAsync(CloseStatusCode code, string reason)
        {
            if (_client && code == CloseStatusCode.ServerError)
            {
                var msg = "ServerError cannot be used.";
                throw new ArgumentException(msg, nameof(code));
            }

            if (!_client && code == CloseStatusCode.MandatoryExtension)
            {
                var msg = "MandatoryExtension cannot be used.";
                throw new ArgumentException(msg, nameof(code));
            }

            if (reason.IsNullOrEmpty())
            {
                closeAsync((ushort)code, string.Empty);
                return;
            }

            if (code == CloseStatusCode.NoStatus)
            {
                var msg = "NoStatus cannot be used.";
                throw new ArgumentException(msg, nameof(code));
            }

            byte[] bytes;
            if (!reason.TryGetUTF8EncodedBytes(out bytes))
            {
                var msg = "It could not be UTF-8-encoded.";
                throw new ArgumentException(msg, nameof(reason));
            }

            if (bytes.Length > 123)
            {
                var msg = "Its size is greater than 123 bytes.";
                throw new ArgumentOutOfRangeException(nameof(reason), msg);
            }

            closeAsync((ushort)code, reason);
        }

        public void Connect()
        {
            string msg;
            if (!checkIfAvailable(true, false, true, false, false, true, out msg))
            {
                Logger.Error(msg);
                error("An error has occurred in connecting.", null);

                return;
            }

            if (connect())
            {
                open();
            }
        }

        public void ConnectAsync()
        {
            string msg;
            if (!checkIfAvailable(true, false, true, false, false, true, out msg))
            {
                Logger.Error(msg);
                error("An error has occurred in connecting.", null);
                return;
            }

            Func<bool> connector = connect;
            Task.Run(() =>
            {
                if (connector())
                {
                    open();
                }
            });
        }

        public bool Ping()
        {
            return ping(EmptyBytes);
        }

        public bool Ping(string message)
        {
            if (message.IsNullOrEmpty())
            {
                return ping(EmptyBytes);
            }

            byte[] bytes;
            if (!message.TryGetUTF8EncodedBytes(out bytes))
            {
                var msg = "It could not be UTF-8-encoded.";
                throw new ArgumentException(msg, nameof(message));
            }

            if (bytes.Length > 125)
            {
                var msg = "Its size is greater than 125 bytes.";
                throw new ArgumentOutOfRangeException(nameof(message), msg);
            }

            return ping(bytes);
        }

        public void Send(byte[] data)
        {
            if (_readyState != WebSocketState.Open)
            {
                var msg = "The current state of the connection is not Open.";
                throw new InvalidOperationException(msg);
            }

            if (data == null)
            {
                throw new ArgumentNullException(nameof(data));
            }

            send(Opcode.Binary, new MemoryStream(data));
        }

        public void Send(FileInfo fileInfo)
        {
            if (_readyState != WebSocketState.Open)
            {
                var msg = "The current state of the connection is not Open.";
                throw new InvalidOperationException(msg);
            }

            if (fileInfo == null)
            {
                throw new ArgumentNullException(nameof(fileInfo));
            }

            if (!fileInfo.Exists)
            {
                var msg = "The file does not exist.";
                throw new ArgumentException(msg, nameof(fileInfo));
            }

            FileStream stream;
            if (!fileInfo.TryOpenRead(out stream))
            {
                var msg = "The file could not be opened.";
                throw new ArgumentException(msg, nameof(fileInfo));
            }

            send(Opcode.Binary, stream);
        }

        public void Send(string data)
        {
            if (_readyState != WebSocketState.Open)
            {
                var msg = "The current state of the connection is not Open.";
                throw new InvalidOperationException(msg);
            }

            if (data == null)
            {
                throw new ArgumentNullException(nameof(data));
            }

            byte[] bytes;
            if (!data.TryGetUTF8EncodedBytes(out bytes))
            {
                var msg = "It could not be UTF-8-encoded.";
                throw new ArgumentException(msg, nameof(data));
            }

            send(Opcode.Text, new MemoryStream(bytes));
        }

        public void Send(Stream stream, int length)
        {
            if (_readyState != WebSocketState.Open)
            {
                var msg = "The current state of the connection is not Open.";
                throw new InvalidOperationException(msg);
            }

            if (stream == null)
            {
                throw new ArgumentNullException(nameof(stream));
            }

            if (!stream.CanRead)
            {
                var msg = "It cannot be read.";
                throw new ArgumentException(msg, nameof(stream));
            }

            if (length < 1)
            {
                var msg = "Less than 1.";
                throw new ArgumentException(msg, nameof(length));
            }

            var bytes = stream.ReadBytes(length);

            var len = bytes.Length;
            if (len == 0)
            {
                var msg = "No data could be read from it.";
                throw new ArgumentException(msg, nameof(stream));
            }

            if (len < length)
            {
                Logger.Warning(
                  string.Format(
                    "Only {0} byte(s) of data could be read from the stream.",
                    len
                  )
                );
            }

            send(Opcode.Binary, new MemoryStream(bytes));
        }

        public void SendAsync(byte[] data, Action<bool> completed = null)
        {
            if (_readyState != WebSocketState.Open)
            {
                var msg = "The current state of the connection is not Open.";
                throw new InvalidOperationException(msg);
            }

            if (data == null)
            {
                throw new ArgumentNullException(nameof(data));
            }

            sendAsync(Opcode.Binary, new MemoryStream(data), completed);
        }

        public void SendAsync(FileInfo fileInfo, Action<bool> completed)
        {
            if (_readyState != WebSocketState.Open)
            {
                var msg = "The current state of the connection is not Open.";
                throw new InvalidOperationException(msg);
            }

            if (fileInfo == null)
            {
                throw new ArgumentNullException(nameof(fileInfo));
            }

            if (!fileInfo.Exists)
            {
                var msg = "The file does not exist.";
                throw new ArgumentException(msg, nameof(fileInfo));
            }

            FileStream stream;
            if (!fileInfo.TryOpenRead(out stream))
            {
                var msg = "The file could not be opened.";
                throw new ArgumentException(msg, nameof(fileInfo));
            }

            sendAsync(Opcode.Binary, stream, completed);
        }

        public void SendAsync(string data, Action<bool> completed)
        {
            if (_readyState != WebSocketState.Open)
            {
                var msg = "The current state of the connection is not Open.";
                throw new InvalidOperationException(msg);
            }

            if (data == null)
            {
                throw new ArgumentNullException(nameof(data));
            }

            byte[] bytes;
            if (!data.TryGetUTF8EncodedBytes(out bytes))
            {
                var msg = "It could not be UTF-8-encoded.";
                throw new ArgumentException(msg, nameof(data));
            }

            sendAsync(Opcode.Text, new MemoryStream(bytes), completed);
        }

        public void SendAsync(Stream stream, int length, Action<bool> completed)
        {
            if (_readyState != WebSocketState.Open)
            {
                var msg = "The current state of the connection is not Open.";
                throw new InvalidOperationException(msg);
            }

            if (stream == null)
            {
                throw new ArgumentNullException(nameof(stream));
            }

            if (!stream.CanRead)
            {
                var msg = "It cannot be read.";
                throw new ArgumentException(msg, nameof(stream));
            }

            if (length < 1)
            {
                var msg = "Less than 1.";
                throw new ArgumentException(msg, nameof(length));
            }

            var bytes = stream.ReadBytes(length);

            var len = bytes.Length;
            if (len == 0)
            {
                var msg = "No data could be read from it.";
                throw new ArgumentException(msg, nameof(stream));
            }

            if (len < length)
            {
                Logger.Warning(
                  string.Format(
                    "Only {0} byte(s) of data could be read from the stream.",
                    len
                  )
                );
            }

            sendAsync(Opcode.Binary, new MemoryStream(bytes), completed);
        }

        public void SetCookie(Cookie cookie)
        {
            string msg;
            if (!checkIfAvailable(true, false, true, false, false, true, out msg))
            {
                Logger.Error(msg);
                error("An error has occurred in setting a cookie.", null);

                return;
            }

            if (cookie == null)
            {
                Logger.Error("'cookie' is null.");
                error("An error has occurred in setting a cookie.", null);

                return;
            }

            lock (_forState)
            {
                if (!checkIfAvailable(true, false, false, true, out msg))
                {
                    Logger.Error(msg);
                    error("An error has occurred in setting a cookie.", null);

                    return;
                }

                lock (CookieCollection.SyncRoot)
                {
                    CookieCollection.SetOrRemove(cookie);
                }
            }
        }

        public void SetCredentials(string username, string password, bool preAuth)
        {
            string msg;
            if (!checkIfAvailable(true, false, true, false, false, true, out msg))
            {
                Logger.Error(msg);
                error("An error has occurred in setting the credentials.", null);

                return;
            }

            if (!checkParametersForSetCredentials(username, password, out msg))
            {
                Logger.Error(msg);
                error("An error has occurred in setting the credentials.", null);

                return;
            }

            lock (_forState)
            {
                if (!checkIfAvailable(true, false, false, true, out msg))
                {
                    Logger.Error(msg);
                    error("An error has occurred in setting the credentials.", null);

                    return;
                }

                if (username.IsNullOrEmpty())
                {
                    Logger.Warning("The credentials are initialized.");
                    Credentials = null;
                    _preAuth = false;

                    return;
                }

                Credentials = new NetworkCredential(username, password, _uri.PathAndQuery);
                _preAuth = preAuth;
            }
        }

        public void SetProxy(string url, string username, string password)
        {
            string msg;
            if (!checkIfAvailable(true, false, true, false, false, true, out msg))
            {
                Logger.Error(msg);
                error("An error has occurred in setting the proxy.", null);

                return;
            }

            if (!checkParametersForSetProxy(url, username, password, out msg))
            {
                Logger.Error(msg);
                error("An error has occurred in setting the proxy.", null);

                return;
            }

            lock (_forState)
            {
                if (!checkIfAvailable(true, false, false, true, out msg))
                {
                    Logger.Error(msg);
                    error("An error has occurred in setting the proxy.", null);

                    return;
                }

                if (url.IsNullOrEmpty())
                {
                    Logger.Warning("The url and credentials for the proxy are initialized.");
                    _proxyUri = null;
                    _proxyCredentials = null;

                    return;
                }

                _proxyUri = new Uri(url);

                if (username.IsNullOrEmpty())
                {
                    Logger.Warning("The credentials for the proxy are initialized.");
                    _proxyCredentials = null;

                    return;
                }

                _proxyCredentials =
                  new NetworkCredential(
                    username, password, string.Format("{0}:{1}", _uri.DnsSafeHost, _uri.Port)
                  );
            }
        }

        void IDisposable.Dispose()
        {
            close(1001, string.Empty);
        }
    }
}